import json
import os
import logging
from typing import Union, Callable

from yadg.parsers import (
    dummy,
    basiccsv,
    qftrace,
    chromtrace,
    flowdata,
    meascsv,
    electrochem,
    masstrace,
    xpstrace,
)
import yadg.dgutils
import yadg.core


def _infer_datagram_handler(parser: str) -> tuple[Callable, str]:
    """
    Helper function to distribute work to `parser`s.

    Add your `parser` here.

    Parameters
    ----------
    parser
        Name of the parser from schema.

    Returns
    -------
    (process, version): tuple[Callable, str]
        A tuple containing the handler function as :class:`(Callable)` and the handler
        version as :class:`(str)`.
    """
    if parser == "chromtrace":
        return chromtrace.process, chromtrace.version
    if parser == "qftrace":
        return qftrace.process, qftrace.version
    if parser == "dummy":
        return dummy.process, dummy.version
    if parser == "basiccsv":
        return basiccsv.process, basiccsv.version
    if parser == "flowdata":
        return flowdata.process, flowdata.version
    if parser == "meascsv":
        return meascsv.process, meascsv.version
    if parser == "electrochem":
        return electrochem.process, electrochem.version
    if parser == "masstrace":
        return masstrace.process, masstrace.version
    if parser == "xpstrace":
        return xpstrace.process, xpstrace.version


def _infer_todo_files(importdict: dict) -> list:
    """
    File enumerator function.

    This function enumerates all paths to be processed by yadg using the "import" key
    within a schema step. Currently, the specification allows for folders, files or paths.

    Parameters
    ----------
    importdict
        A (dict) describing the paths to process. A valid schema has to contain one, and
        only one, of the following keys: ``"folders"``, ``"files"``. Additional keys
        that are processed here are ``"prefix"``, ``"suffix"``, and ``"contains"``.

    Returns
    -------
    todofiles
        A sorted list of paths which match the ``importdict`` spec.
    """
    todofiles = []
    if "folders" in importdict:
        for folder in importdict["folders"]:
            for fn in os.listdir(folder):
                if (
                    fn.startswith(importdict.get("prefix", ""))
                    and fn.endswith(importdict.get("suffix", ""))
                    and importdict.get("contains", "") in fn
                    and not (
                        importdict.get("exclude", False)
                        and importdict.get("exclude", "") in fn
                    )
                ):
                    todofiles.append(os.path.join(folder, fn))
    if "files" in importdict:
        for path in importdict["files"]:
            todofiles.append(path)
    return sorted(todofiles)


def process_schema(schema: Union[list, tuple]) -> dict:
    """
    Main worker function of **yadg**.

    Takes in a validated `schema` as an argument and returns a single annotated
    `datagram` created from the `schema`. It is the job of the user to supply a
    validated `schema`.

    Parameters
    ----------
    schema
        A fully validated `schema`. Use the function :meth:`yadg.core.validators.validate_schema`
        to validate your `schema`.

    Returns
    -------
    datagram: dict
        An unvalidated `datagram`. The `parser`\\ s included in **yadg** should return
        a valid `datagram`; any custom `parser`\\ s might not do so. Use the function
        :meth:`yadg.core.validators.validate_datagram` to validate the resulting `datagram`.
    """
    datagram = {
        "metadata": {
            "provenance": {
                "yadg": yadg.dgutils.get_yadg_metadata(),
            },
            "date": yadg.dgutils.now(asstr=True),
            "input_schema": schema.copy(),
            "datagram_version": yadg.core.spec_datagram.datagram_version,
        },
        "steps": [],
    }
    for step in schema["steps"]:
        metadata = dict()
        common = dict()
        timesteps = list()
        logging.info(f'process_schema: processing step {schema["steps"].index(step)}:')
        handler, parserversion = _infer_datagram_handler(step["parser"])
        metadata["tag"] = step.get("tag", f"{schema['steps'].index(step):02d}")
        metadata["parser"] = {step["parser"]: {"version": parserversion}}
        todofiles = _infer_todo_files(step["import"])
        if len(todofiles) == 0:
            logging.warning(
                f"process_schema: No files processed by step {metadata['tag']}"
            )
        for tf in todofiles:
            logging.debug(f"process_schema: processing item {tf}")
            ts, meta, fulldate = handler(
                tf,
                encoding=step["import"].get("encoding", "utf-8"),
                timezone=schema["metadata"].get("timezone", "localtime"),
                **step.get("parameters", {}),
            )
            ed = step.get("externaldate", {})
            if not fulldate or ed != {}:
                yadg.dgutils.complete_timestamps(
                    ts, tf, ed, schema["metadata"].get("timezone", "localtime")
                )
            assert isinstance(
                ts, list
            ), f"process_schema: Handler for {step['parser']} yields timesteps that are not a enclosed in a `list`."
            timesteps += ts
            assert (
                isinstance(meta, dict) or meta is None
            ), f"process_schema: Handler for {step['parser']} yields metadata that are not a enclosed in a `dict`."
            if meta is not None:
                metadata.update(meta)

        datagram["steps"].append({"metadata": metadata, "data": timesteps})
        if step.get("export", None) is not None:
            with open(step["export"], "w") as ofile:
                json.dump(datagram, ofile, indent=1)
    return datagram



"""
File parser for the LabView CSV files generated using Agilent PNA-L N5320C.

This file format includes a header, with the values of bandwith and averaging,
and three tab-separated columns containing the frequency :math:`f`, and the real
and imaginary parts of the complex reflection coefficient :math:`\\Gamma(f)`.

Timestamps are determined from file name. One trace per file. As the set-up for
which this format was designed always uses the ``S11`` port, the name of the trace
is hard-coded to this value.

.. codeauthor:: Peter Kraus <peter.kraus@empa.ch>
"""

import os
import numpy as np
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

import yadg.dgutils


def process(
    fn: str, encoding: str = "utf-8", timezone: str = "timezone"
) -> tuple[list, dict]:
    """
    VNA reflection trace parser.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    Returns
    -------
    (data, metadata) : tuple[list, None]
        Tuple containing the timesteps, metadata, and common data.
    """

    data = {"fn": str(fn)}
    with open(fn, "r", encoding=encoding) as infile:
        lines = infile.readlines()
    assert (
        len(lines) > 2
    ), f"qftrace: Only {len(lines)-1} points supplied in {fn}; fitting impossible."
    # process header
    bw = [10000.0, 1.0]
    avg = 15
    if ";" in lines[0]:
        items = lines.pop(0).split(";")
        for item in items:
            if item.startswith("BW"):
                bw = tuple_fromstr(item.split("=")[-1].strip())
            if item.startswith("AVG"):
                avg = int(item.split("=")[-1].strip())
    fsbw = bw[0] / avg
    # calculate precision of trace
    data["raw"] = {"traces": {}, "bw": {"n": bw[0], "s": bw[1], "u": "Hz"}, "avg": avg}
    freq = []
    gamma = []
    real = []
    imag = []
    absgamma = []
    for line in lines:
        f, re, im = line.strip().split()
        fn, fs = tuple_fromstr(f)
        fs = max(fs, fsbw)
        ren, res = tuple_fromstr(re)
        imn, ims = tuple_fromstr(im)
        freq.append([fn, fs])
        real.append([ren, res])
        imag.append([imn, ims])
        c = complex(ren, imn)
        gamma.append(c)
        absgamma.append(abs(c))
    temp = {"f": {}, "Re(Γ)": {}, "Im(Γ)": {}}
    freq = [np.array(i) for i in zip(*freq)]
    temp["fvals"], temp["fsigs"] = freq
    temp["gamma"] = np.array(gamma)
    temp["absgamma"] = np.array(absgamma)
    temp["f"]["n"], temp["f"]["s"] = [i.tolist() for i in freq]
    temp["f"]["u"] = "Hz"
    real = [np.array(i) for i in zip(*real)]
    temp["Re(Γ)"]["n"], temp["Re(Γ)"]["s"] = [i.tolist() for i in real]
    temp["Re(Γ)"]["u"] = " "
    imag = [np.array(i) for i in zip(*imag)]
    temp["Im(Γ)"]["n"], temp["Im(Γ)"]["s"] = [i.tolist() for i in imag]
    temp["Im(Γ)"]["u"] = " "
    data["raw"]["traces"]["S11"] = temp
    return [data], None



import os
import pickle
import json
import datetime
import dateutil.parser
import logging
import tzlocal
import zoneinfo
import numpy as np
from typing import Callable, Union
from collections.abc import Iterable


def now(
    asstr: bool = False, tz: datetime.timezone = datetime.timezone.utc
) -> Union[float, str]:
    """
    Wrapper around datetime.now()

    A convenience function for returning the current time as a ISO 8601 or as a unix timestamp.
    """
    dt = datetime.datetime.now(tz=tz)
    if asstr:
        return dt.strftime("%Y-%m-%d %H:%M:%S")
    else:
        return dt.timestamp()


def ole_to_uts(ole_timestamp: float, timezone: str = "UTC") -> float:
    """
    Converts a Microsoft OLE timestamp into a POSIX timestamp.

    The OLE automation date format is a floating point value, counting
    days since midnight 30 December 1899. Hours and minutes are
    represented as fractional days.

    https://devblogs.microsoft.com/oldnewthing/20030905-02/?p=42653

    Parameters
    ----------
    ole_timestamp
        A timestamp in Microsoft OLE format.
    timezone
        String desribing the timezone.

    Returns
    -------
    time: float
        The corresponding Unix timestamp.

    """
    if timezone == "localtime":
        tz = tzlocal.get_localzone()
    else:
        tz = zoneinfo.ZoneInfo(timezone)

    ole_base = datetime.datetime(year=1899, month=12, day=30, tzinfo=tz)
    ole_delta = datetime.timedelta(days=ole_timestamp)
    time = ole_base + ole_delta
    return time.timestamp()


def str_to_uts(
    timestamp: str, format: str = None, timezone: str = "UTC", strict: bool = True
) -> Union[float, None]:
    """
    Converts a string to POSIX timestamp.

    If the optional ``format`` is specified, the ``timestamp`` string is processed
    using the :func:`datetime.datetime.strptime` function; if no ``format`` is
    supplied, an ISO 8601 format is assumed and an attempt to parse using
    :func:`dateutil.parser.parse` is made.

    Parameters
    ----------
    timestamp
        A string containing the timestamp.

    format
        Optional format string for parsing of the ``timestamp``.

    timezone
        Optional timezone of the ``timestamp``. By default, "UTC".

    strict
        Whether to re-raise any parsing errors.

    Returns
    -------
    uts: Union[float, None]
        Returns the POSIX timestamp if successful, otherwise None.

    """
    if timezone == "localtime":
        tz = tzlocal.get_localzone()
    else:
        tz = zoneinfo.ZoneInfo(timezone)

    try:
        if format is None:
            dt = dateutil.parser.parse(timestamp)
        else:
            dt = datetime.datetime.strptime(timestamp, format)

        local_tz = tz if dt.tzinfo is None else dt.tzinfo
        local_dt = dt.replace(tzinfo=local_tz)
        utc_dt = local_dt.astimezone(datetime.timezone.utc)
        return utc_dt.timestamp()
    except (dateutil.parser.ParserError, ValueError) as e:
        if strict:
            raise e
        else:
            logging.info(
                f"str_to_uts: Parsing timestamp '{timestamp}' with "
                f"format '{format}' was not successful."
            )
            return None


def infer_timestamp_from(
    headers: list = None, spec: dict = None, timezone: str = "UTC"
) -> tuple[list, Callable, bool]:
    """
    Convenience function for timestamping

    Given a set of headers, and an optional specification, return an array containing
    column indices from which a timestamp in a given row can be computed, as well as the
    function which will compute the timestamp given the returned array.

    Parameters
    ----------
    headers
        An array of strings. If `spec` is not supplied, must contain either "uts"
        :class:`(float)` or "timestep" :class:`(str)` (conforming to ISO 8601).

    spec
        A specification of timestamp elements with associated column indices and
        optional formats. Currently accepted combinations of keys are: "uts"; "timestamp";
        "date" and / or "time".

    tz
        Timezone to use for conversion. By default, UTC is used.

    Returns
    -------
    (datecolumns, datefunc, fulldate): tuple[list, Callable, bool]
        A tuple containing a list of indices of columns, a Callable to which the
        columns have to be passed to obtain a uts timestamp, and whether the determined
        timestamp is full or partial.

    """

    if spec is not None:
        if "uts" in spec:
            return [spec["uts"].get("index", None)], float, True
        if "timestamp" in spec:

            def retfunc(value):
                return str_to_uts(
                    value, spec["timestamp"].get("format", None), timezone=timezone
                )

            return [spec["timestamp"].get("index", None)], retfunc, True
        if "date" in spec or "time" in spec:
            specdict = {
                "date": datetime.datetime.fromtimestamp(0).timestamp,
                "time": datetime.datetime.fromtimestamp(0).timestamp,
            }
            cols = [None, None]
            if "date" in spec:

                def datefn(value):
                    return str_to_uts(
                        value, spec["date"].get("format", None), timezone=timezone
                    )

                cols[0] = spec["date"].get("index", None)
                specdict["date"] = datefn

            if "time" in spec:
                if "format" in spec["time"]:

                    def timefn(value):
                        t = datetime.datetime.strptime(value, spec["time"]["format"])
                        td = datetime.timedelta(
                            hours=t.hour,
                            minutes=t.minute,
                            seconds=t.second,
                            microseconds=t.microsecond,
                        )
                        return td.total_seconds()

                    cols[1] = spec["time"].get("index", None)
                else:
                    logging.debug(
                        "dateutils: Assuming specified column containing the time is in ISO 8601 format"
                    )

                    def timefn(value):
                        t = datetime.time.fromisoformat(value)
                        td = datetime.timedelta(
                            hours=t.hour,
                            minutes=t.minute,
                            seconds=t.second,
                            microseconds=t.microsecond,
                        )
                        return td.total_seconds()

                    cols[1] = spec["time"].get("index", None)
                specdict["time"] = timefn
            if cols[0] is None:
                return [cols[1]], specdict["time"], False
            elif cols[1] is None:
                return [cols[0]], specdict["date"], False
            else:

                def retfn(date, time):
                    return specdict["date"](date) + specdict["time"](time)

                return cols, retfn, True
    elif "uts" in headers:
        logging.info(
            "dateutils: No timestamp spec provided, assuming column 'uts' is a valid unix timestamp"
        )
        return [headers.index("uts")], float, True
    elif "timestamp" in headers:
        logging.info(
            "dateutils: No timestamp spec provided, assuming column 'timestamp' is a valid ISO 8601 timestamp"
        )

        def retfunc(value):
            return str_to_uts(value, timezone=timezone)

        return [headers.index("timestamp")], retfunc, True
    else:
        assert False, "dateutils: A valid timestamp could not be deduced."


def complete_timestamps(
    timesteps: list, fn: str, spec: dict = {}, timezone: str = "UTC"
) -> None:
    """
    Timestamp completing function.

    This function allows for completing or overriding the uts timestamps determined by
    the individual parsers. **yadg** enters this function for any parser which does not
    return a full timestamp, as well as if the ``externaldate`` specification is
    specified by the user.

    The ``externaldate`` specification is as follows:

    .. code-block:: yaml

        externaldate:
          from:
            file:
                path:    !!str     # path of the external date file
                type:    !!str     # type of the external date file
                match:   !!str     # string to match timestamps with
            isostring:   !!str     # string offset parsed with dateutil
            utsoffset:   !!float   # float offset
            filename:
                format:  !!str     # strptime-style format
                len:     !!int     # length of filename string to match
          mode:          !!str     # "add" or "replace" timestamp

    The ``"from"`` key specifies how an external timestamp is created. Only one entry in
    ``"from"`` is permitted. By default, this entry is:

    .. code-block:: yaml

        externaldate:
          from:
            filename:
              format: "%Y-%m-%d-%H-%M-%S"
              len: 19

    Which means the code will attempt to deduce the timestamp from the path of the
    processed file (``fn``), using the first 19 characters of the base filename
    according to the above format (eg. "2021-12-31-13-45-00").

    If ``"file"`` is specified, the handling of timestamps is handed off to
    :func:`timestamps_from_file`.

    The ``"mode"`` key specifies whether the offsets determined in this function are
    added to the current timestamps (eg. date offset being added to time) or whether
    they should replace the existing timestamps completely.

    As a measure of last resort, the ``mtime`` of the ``fn`` is used. ``mtime`` is
    preferred to ``ctime``, as the former has a more consistent cross-platform
    behaviour.

    Parameters
    ----------
    timesteps
        A list of timesteps generated from a single file, ``fn``.

    fn
        Filename used to create ``timesteps``.

    spec
        ``externaldate`` specification part of the `schema`.

    timezone
        Timezone, defaults to "UTC".

    """

    defaultmethod = {"filename": {"format": "%Y-%m-%d-%H-%M-%S", "len": 19}}

    replace = spec.get("mode", "add") == "replace"
    methods = spec.get("from", defaultmethod)

    delta = None

    if "file" in methods:
        method = methods["file"]
        delta = timestamps_from_file(method["path"], method["type"], timezone)
    elif "isostring" in methods:
        delta = str_to_uts(methods["isostring"], None, timezone, True)
    elif "utsoffset" in methods:
        delta = float(methods["utsoffset"])
    elif "filename" in methods:
        method = methods["filename"]
        dirname, basename = os.path.split(fn)
        filename, ext = os.path.splitext(basename)
        ls = method.get("len", len(filename))
        delta = str_to_uts(filename[:ls], method["format"], timezone, False)

    if delta is None:
        logging.info("complete_timestamps: Timestamp completion failed. Using mtime.")
        delta = os.path.getmtime(fn)

    if isinstance(delta, float):
        delta = [delta] * len(timesteps)

    assert len(delta) == len(timesteps)

    for ts in timesteps:
        if replace:
            ts["uts"] = delta.pop(0)
        else:
            ts["uts"] = ts.get("uts", 0.0) + delta.pop(0)


def timestamps_from_file(path: str, type: str, timezone: str = "UTC") -> list[float]:
    """
    Load timestamps from file.

    This function enables loading timestamps from file. The currently supported
    file formats include ``json`` and ``pkl``, which must contain a top-level
    :class:`(Iterable)` of :class:`(str)` or :class:`(float)`-like objects.

    Parameters
    ----------
    path
        Location of the external file.

    type
        Type of the external file. Currently, ``"json", "pkl"`` are supported.

    timezone
        An optional timezone string, defaults to "UTC"

    Returns
    -------
    parseddata: list[float]
        A list of POSIX timestamps.

    """
    assert os.path.exists(path), f"timestamps_from_file: path '{path}' doesn't exist."
    assert os.path.isfile(path), f"timestamps_from_file: Path '{path}' is not a file."

    if type == "pkl":
        logging.info(f"timestamps_from_file: loading '{path}' as pickle.")
        with open(path, "rb") as infile:
            data = pickle.load(infile)
    elif type == "json":
        logging.info(f"timestamps_from_file: loading '{path}' as json.")
        with open(path, "r") as infile:
            data = json.load(infile)
    elif type == "agilent.log":
        logging.critical(f"timestamps_from_file: type 'agilent.log' not yet supported.")
        data = None

    if data is None:
        return data
    elif isinstance(data, Iterable):
        parseddata = []
        for i in data:
            if isinstance(i, str):
                delta = str_to_uts(i, None, timezone, True)
                parseddata.append(delta)
            elif isinstance(i, (int, float, np.number)):
                parseddata.append(float(i))
        return parseddata



"""Processing of Quadstar 32-bit scan analog data.

The [sac2dat.c code from Dr. Moritz Bubek](https://www.bubek.org/sac2dat.php)
was a really useful stepping stone for this Python adaptation.

Pretty much the entire file format has been reverse engineered. There
are still one or two unknown fields.

File Structure of `.sac` Files
``````````````````````````````

.. code-block:: python

    0x00 "data_index"
    0x02 "software_id"
    0x06 "version_major"
    0x07 "version_minor"
    0x08 "second"
    0x09 "minute"
    0x0a "hour"
    0x0b "day"
    0x0c "month"
    0x0d "year"
    0x0f "author"
    0x64 "n_timesteps"
    0x68 "n_traces"
    0x6a "timestep_length"
    ...
    # Not sure what sits from 0x6e to 0xc2.
    ...
    0xc2 "uts_base_s"
    0xc6 "uts_base_ms"
    # Trace header. Read these 9 bytes for every trace (n_traces).
    0xc8 + (n * 0x09) "type"
    0xc9 + (n * 0x09) "info_position"
    0xcd + (n * 0x09) "data_position"
    ...
    # Trace info. Read these 137 bytes for every trace where type != 0x11.
    info_position + 0x00 "data_format"
    info_position + 0x02 "y_title"
    info_position + 0x0f "y_unit"
    info_position + 0x1d "x_title"
    info_position + 0x2a "x_unit"
    info_position + 0x38 "comment"
    info_position + 0x7a "first_mass"
    info_position + 0x7e "scan_width"
    info_position + 0x80 "values_per_mass"
    info_position + 0x81 "zoom_start"
    info_position + 0x85 "zoom_end"
    ...
    # UTS offset. Read these 6 bytes for every timestep (n_timesteps).
    0xc2 + (n * timestep_length) "uts_offset_s"
    0xc6 + (n * timestep_length) "uts_offset_ms"
    # Read everything remaining below for every timestep and every trace
    # where type != 0x11.
    data_position + (n * timestep_length) + 0x00 "n_datapoints"
    data_position + (n * timestep_length) + 0x04 "data_range"
    # Datapoints. Read these 4 bytes (scan_width * values_per_mass)
    # times.
    data_position + (n * timestep_length) + 0x06 "datapoints"
    ...


Structure of Parsed Timesteps
`````````````````````````````

.. code-block:: yaml

    - fn:  !!str
    - uts: !!float
    - raw:
        traces:
          "{{ trace_number }}":  # number of the trace
            y_title:  !!str      # y-axis label from file
            comment:  !!str      # comment
            fsr:      !!str      # full scale range of detector
            m/z:                 # masses are always in amu
              {n: [!!float, ...], s: [!!float, ...], u: "amu"}
            y:                   # y-axis units from file
              {n: [!!float, ...], s: [!!float, ...], u: !!str}

.. codeauthor:: Nicolas Vetsch <vetschnicolas@gmail.com>
"""
from datetime import datetime
from typing import Any, Union

import numpy as np

# The general header at the top of .sac files.
general_header_dtype = np.dtype(
    [
        ("data_index", "<i2"),
        ("software_id", "<i4"),
        ("version_major", "|u1"),
        ("version_minor", "|u1"),
        ("S", "|u1"),
        ("M", "|u1"),
        ("H", "|u1"),
        ("d", "|u1"),
        ("m", "|u1"),
        ("y", "|u1"),
        ("username", "|S86"),
        ("n_timesteps", "<i4"),
        ("n_traces", "<i2"),
        ("timestep_length", "<i4"),
    ]
)


trace_header_dtype = np.dtype(
    [
        ("type", "|u1"),
        ("info_position", "<i4"),
        ("data_position", "<i4"),
    ]
)


trace_info_dtype = np.dtype(
    [
        ("data_format", "<u2"),
        ("y_title", "|S13"),
        ("y_unit", "|S14"),
        ("x_title", "|S13"),
        ("x_unit", "|S14"),
        ("comment", "|S66"),
        ("first_mass", "<f4"),
        ("scan_width", "<u2"),
        ("values_per_mass", "|u1"),
        ("zoom_start", "<f4"),
        ("zoom_end", "<f4"),
    ]
)


def _read_value(
    data: bytes, offset: int, dtype: np.dtype, encoding: str = "utf-8"
) -> Any:
    """Reads a single value from a buffer at a certain offset.

    Just a handy wrapper for np.frombuffer().

    The read value is converted to a built-in datatype using
    np.dtype.item().

    Parameters
    ----------
    data
        An object that exposes the buffer interface. Here always bytes.

    offset
        Start reading the buffer from this offset (in bytes).

    dtype
        Data-type to read in.

    encoding
        The encoding of the bytes to be converted.

    Returns
    -------
    Any
        The unpacked and converted value from the buffer.

    """
    value = np.frombuffer(data, offset=offset, dtype=dtype, count=1)
    item = value.item()
    if value.dtype.names:
        item = [i.decode(encoding) if isinstance(i, bytes) else i for i in item]
        return dict(zip(value.dtype.names, item))
    return item.decode(encoding) if isinstance(item, bytes) else item


def _read_values(
    data: bytes, offset: int, dtype: np.dtype, count: int
) -> Union[list, list[dict]]:
    """Reads in multiple values from a buffer starting at offset.

    Just a handy wrapper for np.frombuffer() with count >= 1.

    The read values are converted to a list of built-in datatypes using
    np.ndarray.tolist().

    Parameters
    ----------
    data
        An object that exposes the buffer interface. Here always bytes.

    offset
        Start reading the buffer from this offset (in bytes).

    dtype
        Data-type to read in.

    count
        Number of items to read. -1 means all data in the buffer.

    Returns
    -------
    Any
        The values read from the buffer as specified by the arguments.

    """
    values = np.frombuffer(data, offset=offset, dtype=dtype, count=count)
    if values.dtype.names:
        return [dict(zip(value.dtype.names, value.item())) for value in values]
    # The ndarray.tolist() method converts numpy scalars in ndarrays to
    # built-in python scalars. Thus not just list(values).
    return values.tolist()


def _find_first_data_position(scan_headers: list[dict]) -> int:
    """Finds the data position of the first scan containing any data."""
    for header in scan_headers:
        if header["type"] != 0x11:
            continue
        return header["data_position"]


def process(
    fn: str, encoding: str = "utf-8", timezone: str = "localtime"
) -> tuple[list, dict, None]:
    """Processes a Quadstar 32-bit analog data .sac file.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    Returns
    -------
    (data, metadata, common) : tuple[list, dict, None]
        Tuple containing the timesteps, metadata, and common data.

    """
    with open(fn, "rb") as sac_file:
        sac = sac_file.read()
    meta = _read_value(sac, 0x0000, general_header_dtype)
    uts_base_s = _read_value(sac, 0x00C2, "<u4")
    # The ms part of the timestamps is actually saved as tenths of ms so
    # multiplying this by 0.1 here.
    uts_base_ms = _read_value(sac, 0x00C6, "<u2") * 1e-1
    uts_base = uts_base_s + uts_base_ms * 1e-3
    trace_headers = _read_values(sac, 0x00C8, trace_header_dtype, meta["n_traces"])
    # Find the data position of the first data-containing timestep.
    data_pos_0 = _find_first_data_position(trace_headers)
    timesteps = []
    for n in range(meta["n_timesteps"]):
        ts_offset = n * meta["timestep_length"]
        uts_offset_s = _read_value(sac, data_pos_0 - 0x0006 + ts_offset, "<u4")
        uts_offset_ms = _read_value(sac, data_pos_0 - 0x0002 + ts_offset, "<u2") * 1e-1
        uts_timestamp = uts_base + (uts_offset_s + uts_offset_ms * 1e-3)
        traces = {}
        for trace_number, header in enumerate(trace_headers):
            if header["type"] != 0x11:
                continue
            info = _read_value(sac, header["info_position"], trace_info_dtype)
            # Construct the mass data.
            ndm = info["values_per_mass"]
            m_values, dm = np.linspace(
                info["first_mass"],
                info["first_mass"] + info["scan_width"],
                info["scan_width"] * ndm,
                endpoint=False,
                retstep=True,
            )
            m = {"n": m_values.tolist(), "s": [dm] * len(m_values), "u": info["x_unit"]}
            # Read and construct the y data.
            ts_data_pos = header["data_position"] + ts_offset
            # Determine the detector's full scale range.
            fsr = 10 ** _read_value(sac, ts_data_pos + 0x0004, "<i2")
            # The n_datapoints value at timestep_data_position is
            # sometimes wrong. Calculating this here works, however.
            n_datapoints = info["scan_width"] * info["values_per_mass"]
            y_values = _read_values(sac, ts_data_pos + 0x0006, "<f4", n_datapoints)
            # Once a y_value leaves the FSR it jumps to the maximum
            # of a float32. These values should be NaNs instead.
            y_values = [y if y <= fsr else float("NaN") for y in y_values]
            # TODO: Determine the correct accuracy from fsr. The 32bit
            # ADC is a guess that seems to put the error in the correct
            # order of magnitude.
            sigma_adc = fsr / 2 ** 32
            sigma = []
            # Contributions to neighboring masses.
            for i in range(len(y_values)):
                prev_neighbor = 0 if i < ndm else y_values[i - ndm]
                next_neighbor = 0 if i > len(y_values) - ndm - 1 else y_values[i + ndm]
                # The upper limit on contribution from peak at next
                # integer mass is 50ppm.
                sigma_neighbor = 50e-6 * max(prev_neighbor, next_neighbor)
                sigma.append(max(sigma_adc, sigma_neighbor))
            y = {
                "n": y_values,
                "s": sigma,
                "u": info["y_unit"],
            }
            traces[str(trace_number)] = {
                "y_title": info["y_title"],
                "comment": info["comment"],
                "fsr": f"{fsr:.0e}",
                "m/z": m,
                "y": y,
            }
        timesteps.append({"uts": uts_timestamp, "raw": {"traces": traces}, "fn": fn})
    version = str(meta["version_major"]) + "." + str(meta["version_minor"])
    metadata = {
        "params": {
            "software_id": meta["software_id"],
            "version": version,
            "username": meta["username"],
        }
    }
    return timesteps, metadata



from .helpers import get_yadg_metadata
from .dateutils import now, infer_timestamp_from, ole_to_uts, complete_timestamps
from .calib import calib_handler
from .utils import update_object, schema_from_preset
from .btools import read_value
from .pintutils import sanitize_units, ureg



schema_version = "4.0.rc1"

calib = {
    "type": dict,
    "one": {
        "linear": {
            "type": dict,
            "any": {"slope": {"type": float}, "intercept": {"type": float}},
        },
        "inverse": {
            "type": dict,
            "any": {"slope": {"type": float}, "intercept": {"type": float}},
        },
        "poly": {"type": dict, "each": {"type": float}},
        "polynomial": {"type": dict, "each": {"type": float}},
    },
    "any": {
        "atol": {"type": float},
        "rtol": {"type": float},
        "forcezero": {"type": bool},
    },
}

peakdetect = {
    "type": dict,
    "any": {
        "window": {"type": int},
        "polyorder": {"type": int},
        "prominence": {"type": float},
        "threshold": {"type": float},
    },
}

detectors = {
    "type": dict,
    "each": {
        "type": dict,
        "all": {"id": {"type": int}, "peakdetect": peakdetect},
        "any": {"prefer": {"type": bool}},
    },
}

species = {
    "type": dict,
    "each": {
        "type": dict,
        "each": {
            "type": dict,
            "all": {"l": {"type": float}, "r": {"type": float}},
            "any": {"calib": calib},
        },
    },
}

convert = {
    "type": dict,
    "each": {
        "type": dict,
        "each": {"type": dict, "any": {"calib": calib, "fraction": {"type": float}}},
        "any": {"unit": {"type": str}},
    },
}

schema_step = {
    "type": dict,
    "all": {
        "parser": {
            "type": str,
            "one": [
                "dummy",
                "basiccsv",
                "chromtrace",
                "qftrace",
                "flowdata",
                "meascsv",
                "electrochem",
                "masstrace",
                "xpstrace",
            ],
        },
        "import": {
            "type": dict,
            "one": {"files": {"type": list}, "folders": {"type": list}},
            "any": {
                "prefix": {"type": str},
                "suffix": {"type": str},
                "contains": {"type": str},
                "exclude": {"type": str},
                "encoding": {"type": str},
            },
        },
    },
    "any": {
        "tag": {"type": str},
        "export": {"type": str},
        "parameters": {"type": dict, "any": {}, "allow": True},
        "externaldate": {
            "type": dict,
            "any": {
                "from": {
                    "type": dict,
                    "one": {
                        "file": {
                            "type": dict,
                            "all": {"path": {"type": str}, "type": {"type": str}},
                            "any": {"match": {"type": str}},
                        },
                        "isostring": {"type": str},
                        "utsoffset": {"type": float},
                        "filename": {
                            "type": dict,
                            "all": {"format": {"type": str}},
                            "any": {"len": {"type": int}},
                        },
                    },
                },
                "mode": {"type": str, "one": ["replace", "add"]},
            },
        },
    },
}

timestamp = {"type": dict, "any": {"index": {"type": int}, "format": {"type": str}}}

# general parameters
schema_step["any"]["parameters"]["any"].update(
    {
        "atol": {"type": float},
        "rtol": {"type": float},
        "sigma": {
            "type": dict,
            "each": {
                "type": dict,
                "any": {"atol": {"type": float}, "rtol": {"type": float}},
            },
        },
        "calfile": {"type": str},
    }
)

# basiccsv parameters
schema_step["any"]["parameters"]["any"].update(
    {
        "sep": {"type": str},
        "units": {"type": dict, "each": {"type": str}},
        "timestamp": {
            "type": dict,
            "any": {
                "timestamp": timestamp,
                "uts": timestamp,
                "date": timestamp,
                "time": timestamp,
            },
        },
        "convert": convert,
    }
)

# chromtrace parameters
schema_step["any"]["parameters"]["any"].update(
    {
        "tracetype": {
            "type": str,
            "one": [],
        },
        "species": species,
        "detectors": detectors,
    }
)
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("ezchrom.asc")
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("fusion.json")
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("fusion.zip")
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("agilent.ch")
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("agilent.dx")
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("agilent.csv")

# qftrace parameters
schema_step["any"]["parameters"]["any"].update(
    {
        "method": {"type": str, "one": ["naive", "lorentz", "kajfez"]},
        "height": {"type": float},
        "distance": {"type": float},
        "cutoff": {"type": float},
        "threshold": {"type": float},
    }
)
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("labview.csv")

# masstrace parameters
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("quadstar.sac")

# xpstrace parameters
schema_step["any"]["parameters"]["any"]["tracetype"]["one"].append("phi.spe")

# flowdata parameters
schema_step["any"]["parameters"]["any"].update(
    {
        "filetype": {
            "type": str,
            "one": [],
        },
    }
)
schema_step["any"]["parameters"]["any"]["filetype"]["one"].append("drycal")
schema_step["any"]["parameters"]["any"]["filetype"]["one"].append("drycal.csv")
schema_step["any"]["parameters"]["any"]["filetype"]["one"].append("drycal.rtf")
schema_step["any"]["parameters"]["any"]["filetype"]["one"].append("drycal.txt")

schema = {
    "type": dict,
    "all": {
        "metadata": {
            "type": dict,
            "all": {
                "provenance": {"type": (dict, str)},
                "schema_version": {"type": str},
            },
            "any": {"timezone": {"type": str}},
        },
        "steps": {"type": list, "each": schema_step},
    },
}

# electrochem parameters
schema_step["any"]["parameters"]["any"]["filetype"]["one"].append("eclab.mpt")
schema_step["any"]["parameters"]["any"]["filetype"]["one"].append("eclab.mpr")



#!/usr/bin/env python
import argparse
import os
import logging
from importlib import metadata

import yadg.subcommands


def set_loglevel(delta: int):
    loglevel = min(max(30 - (10 * delta), 10), 50)
    logging.basicConfig(level=loglevel)
    logging.debug(f"loglevel set to '{logging._levelToName[loglevel]}'")


def run_with_arguments():
    """
    Main execution function.

    This is the function executed when **yadg** is launched using the executable.
    The function has three subcommands:

    - ``process``: processes a given schema into a datagram.
    - ``update``: updates a given schema or datagram to the current version.
    - ``preset``: creates a schema from a preset file and a target folder.

    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version=f'%(prog)s version {metadata.version("yadg")}',
    )

    verbose = argparse.ArgumentParser(add_help=False)

    for p in [parser, verbose]:
        p.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="Increase verbosity by one level.",
        )
        p.add_argument(
            "-q",
            "--quiet",
            action="count",
            default=0,
            help="Decrease verbosity by one level.",
        )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    process = subparsers.add_parser("process")
    process.add_argument(
        "infile",
        help="File containing the schema to be processed by yadg.",
        default=None,
    )
    process.add_argument(
        "outfile",
        nargs="?",
        help="Output file to save the created datagram to.",
        default="datagram.json",
    )
    process.add_argument(
        "--ignore-file-errors",
        dest="permissive",
        action="store_true",
        help="Ignore file opening errors while processing schemafile",
        default=False,
    )
    process.set_defaults(func=yadg.subcommands.process)

    update = subparsers.add_parser("update")
    update.add_argument(
        "type",
        choices=["schema", "datagram"],
        help="Specify whether a schema or a datagram is to be expected.",
    )
    update.add_argument(
        "infile",
        help="The file containing the schema/datagram in the old format.",
    )
    update.add_argument(
        "outfile",
        nargs="?",
        help="Output file to save the updated object to.",
        default=None,
    )
    update.set_defaults(func=yadg.subcommands.update)

    preset = subparsers.add_parser("preset")
    preset.add_argument(
        "preset",
        help="Specify a schema template from a 'preset'.",
    )
    preset.add_argument(
        "folder",
        help="Specify the 'folder' on which to apply the 'preset'.",
    )
    preset.add_argument(
        "outfile",
        nargs="?",
        help=(
            "Output file to save the created schema file. Default is 'schema.json'. "
            "If '--process' is specified, the created datagram file will be saved "
            "instead. Default in that case is 'datagram.json'."
        ),
        default=None,
    )
    preset.add_argument(
        "--process",
        "-p",
        action="store_true",
        help="Immediately process the schema created from the preset.",
        default=False,
    )
    preset.set_defaults(func=yadg.subcommands.preset)

    # parse subparser args
    args, extras = parser.parse_known_args()
    # parse extras for verbose tags
    args, extras = verbose.parse_known_args(extras, args)

    set_loglevel(args.verbose - args.quiet)
    if "func" in args:
        args.func(args)



"""
Trace-pruning routines.

To reduce the impact of baseline noise and  speed up the fitting of :math:`Q_0` to
the :math:`\\Gamma(f)` trace, only smaller sections of the trace are retained around
each minimum. Two routines for trace-pruning are provided:

- a height cutoff-based prune in :func:`yadg.parsers.qftrace.prune.cutoff`
- a gradient threshold-based prune in :func:`yadg.parsers.qftrace.prune.threshold`

.. codeauthor:: Peter Kraus <peter.kraus@empa.ch>
"""

import numpy as np
from uncertainties import unumpy


def cutoff(p0: int, absgamma: np.ndarray, cutoff: float) -> tuple[int, int]:
    """
    Cutoff-based prune.

    Parameters
    ----------
    p0
        Index of the minimum of ``absgamma``.
    absgamma
        The :math:`|\\Gamma(f)|` data.
    cutoff
        Normalised height above which the trace is cut off.

    Returns
    -------
    tuple[int, int]
        A tuple of the left and right index of the trimmed trace.
    """

    min_v = absgamma[p0]
    max_v = absgamma.max()
    norm = absgamma - (min_v / (max_v / min_v))
    for l in range(p0 - 1):
        li = p0 - l
        if norm[li] <= cutoff:
            pass
        else:
            break
    for r in range(len(absgamma) - p0):
        ri = p0 + r
        if norm[ri] <= cutoff:
            pass
        else:
            break
    return li + 1, ri - 1


def gradient(p0: int, absgamma: np.ndarray, threshold: float) -> tuple[int, int]:
    """
    Gradient-based prune.

    Parameters
    ----------
    p0
        Index of the minimum of ``absgamma``.
    absgamma
        The :math:`|\\Gamma(f)|` data.
    threshold
        Threshold of the gradient below which the trace is cut off.

    Returns
    -------
    tuple[int, int]
        A tuple of the left and right index of the trimmed trace.
    """

    grad = np.gradient(absgamma)
    for l in range(p0 - 1):
        li = p0 - l
        if abs(grad[li]) > threshold or l < 100:
            pass
        else:
            break
    for r in range(absgamma.size - p0):
        ri = p0 + r
        if abs(grad[ri]) > threshold or r < 100:
            pass
        else:
            break
    return li + 1, ri - 1



"""
.. _parsers_flowdata_formats:

List of supported file formats:

 - DryCal log file text output (txt) :mod:`yadg.parsers.flowdata.drycal`
 - DryCal log file tabulated output (csv) :mod:`yadg.parsers.flowdata.drycal`
 - DryCal log file document file (rtf) :mod:`yadg.parsers.flowdata.drycal`
"""
from .main import process, version



import logging
import json
import uncertainties as uc
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
import yadg.dgutils
from typing import Callable

version = "4.0.0"


def process_row(
    headers: list,
    items: list,
    units: dict,
    datefunc: Callable,
    datecolumns: list,
    calib: dict = {},
) -> dict:
    """
    A function that processes a row of a table.

    This is the main worker function of ``basiccsv``, but can be re-used by any other
    parser that needs to process tabular data.

    .. _processing_convert:

    This function processes the ``"calib"`` parameter, which should be a :class:`(dict)`
    in the following format:

    .. code-block:: yaml

      - new_name:     !!str    # derived entry name
        - old_name:   !!str    # raw header name
          - calib: {}          # calibration specification
          fraction:   !!float  # coefficient for linear combinations of old_name
        unit:         !!str    # unit of new_name

    The syntax of the calibration specification is detailed in
    :func:`yadg.dgutils.calib.calib_handler`.

    Parameters
    ----------
    headers
        A list of headers of the table.

    items
        A list of values corresponding to the headers. Must be the same length as headers.

    units
        A dict for looking up the units corresponding to a certain header.

    datefunc
        A function that will generate ``uts`` given a list of values.

    datecolumns
        Column indices that need to be passed to ``datefunc`` to generate uts.

    calib
        Specification for converting raw data in ``headers`` and ``items`` to other
        quantities. Arbitrary linear combinations of ``headers`` are possible. See
        :ref:`the above section<processing_convert>` for the specification.

    Returns
    -------
    element: dict
        A result dictionary, containing the keys ``"uts"`` with a timestamp,
        ``"raw"`` for all raw data present in the headers, and ``"derived"``
        for any data processes via ``calib``.

    """
    assert len(headers) == len(items), (
        f"process_row: Length mismatch between provided headers: "
        f"{headers} and  provided items: {items}."
    )

    assert all([key in units for key in headers]), (
        f"process_row: Not all entries in provided 'headers' are present "
        f"in provided 'units': {headers} vs {units.keys()}"
    )

    raw = dict()
    der = dict()
    element = {"raw": dict()}
    columns = [column.strip() for column in items]

    # Process raw data, assign sigma and units
    element["uts"] = datefunc(*[columns[i] for i in datecolumns])
    for header in headers:
        ci = headers.index(header)
        if ci in datecolumns:
            continue
        try:
            val, sig = tuple_fromstr(columns[ci])
            unit = units[header]
            element["raw"][header] = {"n": val, "s": sig, "u": unit}
            raw[header] = (val, sig)
        except ValueError:
            element["raw"][header] = columns[ci]

    # Process calib
    for newk, spec in calib.items():
        y = uc.ufloat(0, 0)
        for oldk, v in spec.items():
            if oldk in der:
                dy = yadg.dgutils.calib_handler(
                    der[oldk],
                    v.get("calib", None),
                )
                y += dy * v.get("fraction", 1.0)
            elif oldk in raw:
                dy = yadg.dgutils.calib_handler(
                    uc.ufloat(*raw[oldk]),
                    v.get("calib", None),
                )
                y += dy * v.get("fraction", 1.0)
            elif oldk == "unit":
                pass
            else:
                logging.warning(
                    f"process_row: Supplied key '{oldk}' is neither a 'raw' nor a 'derived' key."
                )
        if "derived" not in element:
            element["derived"] = dict()
        element["derived"][newk] = {"n": y.n, "s": y.s, "u": spec.get("unit", " ")}
        der[newk] = y
    return element


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    sep: str = ",",
    units: dict = None,
    timestamp: dict = None,
    convert: dict = None,
    calfile: str = None,
) -> tuple[list, dict, bool]:
    """
    A basic csv parser.

    This parser processes a csv file. The header of the csv file consists of one or two
    lines, with the column headers in the first line and the units in the second. The
    parser also attempts to parse column names to produce a timestamp, and save all other
    columns as floats or strings.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    sep
        Separator to use. Default is "," for csv.

    units
        Column-specific unit specification. If present, even if empty, 2nd line is
        treated as data. If omitted, 2nd line is treated as units.

    timestamp
        Specification for timestamping. Allowed keys are ``"date"``, ``"time"``,
        ``"timestamp"``, ``"uts"``. The entries can be ``"index"`` :class:`(list[int])`,
        containing the column indices, and ``"format"`` :class:`(str)` with the format
        string to be used to parse the date. See :func:`yadg.dgutils.dateutils.infer_timestamp_from`
        for more info.

    convert
        Specification for column conversion. The key of each entry will form a new
        datapoint in the ``"derived"`` :class:`(dict)` of a timestep, including the
        option to specify linear combinations. See :ref:`here<processing_convert>` for
        more info.

    calfile
        ``convert``-like functionality specified in a json file.

    Returns
    -------
    (data, metadata, fulldate): tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. No metadata is
        returned by the basiccsv parser. The full date might not be returned, eg.
        when only time is specified in columns.

    """
    # Process calfile and convert into calib
    if calfile is not None:
        with open(calfile, "r") as infile:
            calib = json.load(infile)
    else:
        calib = {}
    if convert is not None:
        calib.update(convert)

    # Load file, extract headers and get timestamping function
    with open(fn, "r", encoding=encoding) as infile:
        # This decode/encode is done to account for some csv files that have a BOM
        # at the beginning of each line.
        lines = [i.encode().decode(encoding) for i in infile.readlines()]
    assert len(lines) >= 2
    headers = [header.strip() for header in lines[0].split(sep)]
    datecolumns, datefunc, fulldate = yadg.dgutils.infer_timestamp_from(
        headers=headers, spec=timestamp, timezone=timezone
    )

    # Populate units
    if units is None:
        units = {}
        _units = [column.strip() for column in lines[1].split(sep)]
        for header in headers:
            units[header] = _units.pop(0)
        si = 2
    else:
        for header in headers:
            if header not in units:
                logging.warning(f"Using implicit dimensionless unit '-' for {header}.")
                units[header] = " "
            elif units[header] == "":
                units[header] = " "
        si = 1

    yadg.dgutils.sanitize_units(units)

    # Process rows
    data = []
    for line in lines[si:]:
        element = process_row(
            headers, line.split(sep), units, datefunc, datecolumns, calib=calib
        )
        element["fn"] = str(fn)
        data.append(element)
    return data, None, fulldate



import logging
import json

from yadg.parsers.flowdata import drycal
import yadg.dgutils

version = "4.0.0"


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    filetype: str = "drycal",
    convert: dict = None,
    calfile: str = None,
    date: str = None,
) -> tuple[list, dict, bool]:
    """
    Flow meter data processor

    This parser processes flow meter data.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    filetype
        Whether a rtf, csv, or txt file is to be expected. When `None`, the suffix of
        the file is used to determine the file type.

    convert
        Specification for column conversion. The `key` of each entry will form a new
        datapoint in the ``"derived"`` :class:`(dict)` of a timestep. The elements within
        each entry must either be one of the ``"header"`` fields, or ``"unit"`` :class:`(str)`
        specification. See :ref:`processing convert<processing_convert>` for more info.

    calfile
        ``convert``-like functionality specified in a json file.

    date
        An optional date argument, required for parsing DryCal files. Otherwise the date
        is parsed from ``fn``.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. Whether full date
        is returned depends on the file parser.

    """
    if calfile is not None:
        with open(calfile, "r") as infile:
            calib = json.load(infile)
    else:
        calib = {}
    if convert is not None:
        calib.update(convert)

    metadata = {}

    if filetype.startswith("drycal"):
        fulldate = False

        if "flow" not in calib:
            logging.info("flowdata: adding default 'DryCal' -> 'flow' conversion")
            calib["flow"] = {"DryCal": {"calib": {"linear": {"slope": 1.0}}}}

        if filetype.endswith(".rtf") or fn.endswith("rtf"):
            ts, meta = drycal.rtf(fn, encoding, timezone, calib)
        elif filetype.endswith(".csv") or fn.endswith("csv"):
            ts, meta = drycal.sep(fn, ",", encoding, timezone, calib)
        elif filetype.endswith(".txt") or fn.endswith("txt"):
            ts, meta = drycal.sep(fn, "\t", encoding, timezone, calib)

    metadata.update(meta)

    return ts, metadata, fulldate



"""
File parser for Agilent Chemstation Chromtab tabulated data files (csv).

This file format may include more than one timestep in each CSV file. It contains
a header section for each timestep, followed by a detector name, and a sequence of
[X, Y] datapoints.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   None
      sampleid: !!str
      username: None
      version:  None
      valve:    None
      datafile: !!str

Unfortunately, neither ``method`` nor ``version`` are exposed, which is a big weakness
of this file format.

.. codeauthor:: Peter Kraus <peter.kraus.empa.ch>
"""
import numpy as np
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

import yadg.dgutils
from yadg.dgutils.dateutils import str_to_uts


def _process_headers(headers: list, columns: list, timezone: str) -> dict:
    res = {}
    assert len(headers) == len(
        columns
    ), f"chromtab: The number of headers and columns do not match."
    assert "Date Acquired" in headers, "chromtab: Cannot infer date."
    res["uts"] = str_to_uts(
        columns[headers.index("Date Acquired")].strip(),
        format="%d %b %Y %H:%M",
        timezone=timezone,
    )
    fn = ""
    if "Path" in headers:
        fn += columns[headers.index("Path")]
    if "File" in headers:
        fn += columns[headers.index("File")]
    res["datafile"] = fn
    if "Sample" in headers:
        res["sampleid"] = columns[headers.index("Sample")]
    return res


def _to_trace(tx, ty):
    xsn, xss = [np.array(x) * 60 for x in zip(*tx)]
    xs = [xsn, xss]
    ysn, yss = [np.array(y) for y in zip(*ty)]
    ys = [ysn, yss]
    trace = {
        "t": {"n": xsn.tolist(), "s": xss.tolist(), "u": "s"},
        "y": {"n": ysn.tolist(), "s": yss.tolist(), "u": " "},
        "data": [xs, ys],
    }
    return trace


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict]:
    """
    Agilent Chemstation CSV (Chromtab) file parser

    Each file may contain multiple chromatograms per file with multiple traces. Each
    chromatogram starts with a header section, and is followed by each trace, which
    includes a header line and x,y-data.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    (chroms, metadata): tuple[list, dict]
        Standard timesteps & metadata tuple.
    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()
    metadata = {
        "filetype": "chromtab",
        "params": {
            "method": None,
            "valve": None,
            "username": None,
            "version": None,
        },
    }
    chroms = []
    chrom = {"fn": str(fn), "traces": {}}
    tx = []
    ty = []
    for li in range(len(lines)):
        line = lines[li].strip()
        parts = line.strip().split(",")
        if len(parts) > 2:
            if '"Date Acquired"' in parts:
                if tx != [] and ty != []:
                    trace = _to_trace(tx, ty)
                    trace["id"] = len(chrom["traces"])
                    chrom["traces"][detname] = trace
                    tx = []
                    ty = []
                if chrom != {"fn": fn, "traces": {}}:
                    chroms.append(chrom)
                    chrom = {"fn": fn, "traces": {}}
                headers = [p.replace('"', "") for p in parts]
            else:
                columns = [p.replace('"', "") for p in parts]
                ret = _process_headers(headers, columns, timezone)
                chrom["uts"] = ret.pop("uts")
                metadata["params"].update(ret)
        elif len(parts) == 1:
            if tx != [] and ty != []:
                trace = _to_trace(tx, ty)
                trace["id"] = len(chrom["traces"])
                chrom["traces"][detname] = trace
                tx = []
                ty = []
            detname = parts[0].replace('"', "").split("\\")[-1]
        elif len(parts) == 2:
            x, y = [tuple_fromstr(i) for i in parts]
            tx.append(x)
            ty.append(y)
    trace = _to_trace(tx, ty)
    trace["id"] = len(chrom["traces"])
    chrom["traces"][detname] = trace
    chroms.append(chrom)
    return chroms, metadata



"""
File parser for Fusion zip archive format (zip).

This is a wrapper parser which unzips the provided zip file, and then uses
the :mod:`yadg.parsers.chromtrace.fusionjson` parser to parse every data
file present in the archive.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      sampleid: !!str
      username: None
      version:  !!str
      valve:    !!int
      datafile: !!str

.. codeauthor:: Peter Kraus
"""
import zipfile
import tempfile
import os

from .fusionjson import process as processjson


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict]:
    """
    Fusion zip file format.

    The Fusion GC's can export their json formats as a zip archive of a folder
    of jsons. This parser allows for parsing of this zip archive directly,
    without the user having to unzip & move the data.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Not used as the file is binary.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    (chroms, metadata): tuple[list, dict]
        Standard timesteps & metadata tuple.
    """

    zf = zipfile.ZipFile(fn)
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        chroms = []
        meta = {}
        for ffn in sorted(os.listdir(tempdir)):
            ffn = os.path.join(tempdir, ffn)
            if ffn.endswith("fusion-data"):
                _chrom, _meta = processjson(ffn, encoding, timezone)
                for ts in _chrom:
                    ts["fn"] = str(fn)
                    chroms.append(ts)
                if _meta is not None:
                    meta.update(_meta)
                meta["params"]["datafile"] = str(ffn)
    return chroms, meta



"""
File parser for Agilent OpenLab data archive files (DX).

This is a wrapper parser which unzips the provided DX file, and then uses the 
:mod:`yadg.parsers.chromtrace.agilentch` parser to parse every CH file present in
the archive. The IT files in the archive are currently ignored.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      sampleid: !!str
      username: !!str
      version:  !!str
      valve:    None
      datafile: !!str

In addition to the metadata exposed by the CH parser, the ``datafile`` entry
is populated with the corresponding name of the CH file. The ``fn`` entry in each
timestep contains the parent DX file.

.. note::

   Currently the timesteps from multiple CH files (if present) are appended in the
   timesteps array without any further sorting.

.. codeauthor:: Peter Kraus
"""
import zipfile
import tempfile
import os

from .agilentch import process as processch


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict]:
    """
    Agilent OpenLab DX archive parser.

    This is a simple wrapper around the Agilent OpenLab signal trace parser in
    :mod:`yadg.parsers.chromtrace.agilentch`. This wrapper first un-zips the DX
    file into a temporary directory, and then processess all CH files found
    within the archive, concatenating timesteps from multiple files.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Not used as the file is binary.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    (chroms, metadata): tuple[list, dict]
        Standard timesteps & metadata tuple.
    """

    zf = zipfile.ZipFile(fn)
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        chroms = []
        meta = {}
        common = None
        for ffn in os.listdir(tempdir):
            if ffn.endswith("CH"):
                _chrom, _meta = processch(ffn, encoding, timezone)
                for ts in _chrom:
                    ts["fn"] = str(fn)
                    chroms.append(ts)
                if _meta is not None:
                    meta.update(_meta)
                meta["params"]["datafile"] = str(ffn)
    return chroms, meta



from uncertainties import ufloat, UFloat
from typing import Union

_default_calib = {"linear": {"slope": 1.0, "intercept": 0.0}, "atol": 0, "rtol": 1e-3}


def _linear(x: UFloat, calspec: dict) -> UFloat:
    c = calspec.get("intercept", 0.0)
    m = calspec.get("slope", 1.0)
    y = m * x + c
    return y


def _inverse(y: UFloat, calspec: dict) -> UFloat:
    c = calspec.get("intercept", 0.0)
    m = calspec.get("slope", 1.0)
    x = (y - c) / m
    return x


def _poly(x: UFloat, calspec: dict) -> UFloat:
    y = 0
    for k, v in calspec.items():
        if k == "c":
            y += v
        elif k.startswith("c"):
            o = int(k[1:])
            y += v * (x ** o)
    return y


def calib_handler(
    x: Union[float, UFloat], calib: dict = None, atol: float = 0.0, rtol: float = 0.0
) -> UFloat:
    """
    Calibration handling function.

    Returns ``y`` calculated from ``x`` and the other supplied arguments.

    .. _processing_calib:

    The ``"calib"`` parameter is a :class:`(dict)` in the following format:

    .. code-block:: yaml

       - calib:
         - linear:                # y = slope * x + intercept
           intercept:  !!float
           slope:      !!float
         - inverse:               # y = (x - intercept) / slope
           intercept:  !!float
           slope:      !!float
         - polynomial:            # y = sum(cN * x**N)
           c0:         !!float
           ...
           cN:         !!float
         atol:         !!float    # absolute uncertainty of y
         rtol:         !!float    # relative uncertainty of y

    Parameters
    ----------
    x
        The raw value

    calib
        Calibration dictionary, specified using the format
        :ref:`described above<processing_calib>`. If empty, corresponds to
        no transformation (i.e. ``linear`` with ``"slope" = 1``) and a ``rtol = 1e-3``.

    atol
        Optional absolute uncertainty of y, overrides that provided in ``calib``.

    rtol
        Optional relative uncertainty of y, overrides that provided in ``calib``.

    Returns
    -------
    y : UFloat
        The derived value.
    """

    if not isinstance(x, UFloat):
        x = ufloat(x, 0.0)
    if calib is None:
        calib = _default_calib
    if x.n == 0.0 and calib.get("forcezero", True):
        return x
    if "linear" in calib:
        y = _linear(x, calib["linear"])
    elif "inverse" in calib:
        y = _inverse(x, calib["inverse"])
    elif "polynomial" in calib:
        y = _poly(x, calib["polynomial"])
    elif "poly" in calib:
        y = _poly(x, calib["poly"])
    y = ufloat(
        y.n, max(y.s, calib.get("atol", atol), abs(y.n * calib.get("rtol", rtol)))
    )
    return y



import os
import argparse
import logging
import json
import yadg.core
import yadg.dgutils


def process(args: argparse.Namespace) -> None:
    """
    The ``process`` subcommand of **yadg**.

    This function first checks that the supplied ``args.infile`` exists, is a valid
    `schema`, and if yes, proceeds to process the `schema` into a `datagram`. If
    this is successful, the `datagram` is checked for validity and written out
    into ``args.outfile`` (which is `"datagram.json"` by default).
    """
    assert os.path.exists(args.infile) and os.path.isfile(args.infile), (
        f"Supplied schema filename '{args.infile}' does not exist "
        f"or is not a valid file."
    )

    logging.info(f"yadg process: Reading input json from '{args.infile}'.")
    with open(args.infile, "r") as infile:
        schema = json.load(infile)

    logging.debug("yadg process: Validating schema.")
    assert yadg.core.validate_schema(schema, args.permissive)

    logging.debug("yadg process: Processing schema")
    datagram = yadg.core.process_schema(schema)

    logging.info(f"yadg process: Saving datagram to '{args.outfile}'.")
    with open(args.outfile, "w") as ofile:
        json.dump(datagram, ofile, indent=1)


def update(args: argparse.Namespace) -> None:
    """
    The ``update`` subcommand of **yadg**.

    This function requires the ``args.type`` argument, which defines whether the
    object to be updated is a `datagram` or a `schema`. Generally, updating a
    `schema` and then re-processing the raw data is preferred to updating a
    `datagram`.

    If the supplied ``args.infile`` exists, and can be parsed as a json, it will
    be passed to :func:`yadg.dgutils.update_object` to perform the update. If the
    update is successful, the object is checked for validity and written out
    into ``args.outfile`` (which is the ``args.infile`` with  a `".new.json"`
    suffix by default).
    """
    assert args.type in ["datagram", "schema"], (
        f"Specified object type '{args.type}' is not one of " f"['datagram', 'schema']."
    )

    assert os.path.exists(args.infile) and os.path.isfile(args.infile), (
        f"Supplied object filename '{args.infile}' does not exist "
        f"or is not a valid file."
    )

    logging.info(f"yadg update: Reading input json from '{args.infile}'.")
    with open(args.infile) as infile:
        inobj = json.load(infile)

    if args.outfile is None:
        name, ext = os.path.splitext(args.infile)
        args.outfile = f"{name}.new.json"

    logging.info(f"yadg update: Updating old object.")
    outobj = yadg.dgutils.update_object(args.type, inobj)

    logging.info(f"yadg update: Writing new object into '{args.outfile}'.")
    with open(args.outfile, "w") as outfile:
        json.dump(outobj, outfile, indent=1)


def preset(args: argparse.Namespace) -> None:
    """
    The ``preset`` subcommand of **yadg**.

    This function requires the ``args.preset`` and ``args.folder`` arguments.
    If ``args.folder`` is an existing folder and ``args.preset`` a valid `schema`
    file (with appropriately formated ``"import"`` and ``"calfile"`` entries),
    this function will prepend the specified ``args.folder`` to all paths in the
    ``args.preset``, convert them to absolute paths, and save the resulting `schema`
    in the supplied ``args.outfile``.

    Alternatively, if ``args.process`` is specified, the created `schema` will be
    directly processed into a `datagram`, which is then saved in ``args.outfile``.
    """
    assert os.path.exists(args.folder) and os.path.isdir(args.folder), (
        f"Supplied folder path '{args.folder}' does not exist "
        f"or is not a valid folder."
    )

    if not os.path.isabs(args.folder) and not args.process:
        logging.warning(
            f"yadg preset: The provided path '{args.folder}' is a relative path. "
            "The generated schema likely will not work outside current working "
            "directory. "
        )

    assert os.path.exists(args.preset) and os.path.isfile(args.preset), (
        f"Supplied preset path '{args.preset}' does not exist "
        f"or is not a valid file."
    )

    logging.info(f"yadg preset: Reading input json from '{args.preset}'.")
    with open(args.preset) as infile:
        preset = json.load(infile)

    logging.info(f"yadg preset: Validating loaded preset.")
    assert yadg.core.validate_schema(preset, strictfiles=False, strictfolders=False)

    logging.info(f"yadg preset: Creating a schema from preset for '{args.folder}'.")
    schema = yadg.dgutils.schema_from_preset(preset, args.folder)

    logging.info(f"yadg preset: Validating created schema.")
    assert yadg.core.validate_schema(schema)

    if args.process:
        logging.info(f"yadg preset: Processing created schema.")
        datagram = yadg.core.process_schema(schema)
        args.outfile = "datagram.json" if args.outfile is None else args.outfile
        logging.info(f"yadg preset: Saving datagram to '{args.outfile}'.")
        with open(args.outfile, "w") as ofile:
            json.dump(datagram, ofile, indent=1)
    else:
        args.outfile = "schema.json" if args.outfile is None else args.outfile
        logging.info(f"yadg preset: Saving schema to '{args.outfile}'.")
        with open(args.outfile, "w") as ofile:
            json.dump(schema, ofile, indent=1)



"""Implemented technique parameters.

Implemented techniques:
    CA - Chronoamperometry / Chronocoulometry
    CP - Chronopotentiometry
    CV - Cyclic Voltammetry
    GCPL - Galvanostatic Cycling with Potential Limitation
    GEIS - Galvano Electrochemical Impedance Spectroscopy
    LOOP - Loop
    LSV - Linear Sweep Voltammetry
    MB - Modulo Bat
    OCV - Open Circuit Voltage
    PEIS - Potentio Electrochemical Impedance Spectroscopy
    WAIT - Wait
    ZIR - IR compensation (PEIS)

The module also implements resolution determination for parameters of techniques, 
in :func:`get_resolution`.

.. admonition:: TODO

    https://github.com/dgbowl/yadg/issues/10

    Current values of the uncertainties ``"s"`` are hard-coded from VMP-3 values
    of resolutions and accuracies, with ``math.ulp(n)`` as fallback. The values 
    should be device-specific, and the fallback should be eliminated.

"""
import re

import numpy as np
from typing import Union
import bisect
import math
import logging

# Short helper function for constructing params.
def _prepend_ns(settings: list[str], params: list) -> list[str]:
    """Prepends the 'Ns' parameter to the parameters if present.

    Parameters
    ----------
    settings
        The list of settings from the start of an .mpt or .mps file.
    params
        A list containing the first part of the technique parameter
        names. The Ns is prepended to this if present in the settings.

    Returns
    -------
    list[str]
        The 'Ns'-prepended parameters or the unchanged parameters.

    """
    ns_match = re.search(r"^Ns.+", "\n".join(settings))
    if ns_match:
        return ["Ns"] + params
    return params


################# Chronoamperometry / Chronocoulometry #################
def _ca_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the CA technique."""
    params = [
        "Ei",
        "Ei_vs",
        "ti",
        "Imax",
        "Imax_unit",
        "Imin",
        "Imin_unit",
        "dQM",
        "dQM_unit",
        "record",
        "dI",
        "dI_unit",
        "dQ",
        "dQ_unit",
        "dt",
        "dta",
        "E_range_min",
        "E_range_max",
        "I_range",
        "I_range_min",
        "I_range_max",
        "I_range_init",
        "bandwidth",
        "goto_Ns",
        "nc_cycles",
    ]
    return _prepend_ns(settings, params)


_ca_params_dtype = np.dtype(
    [
        ("Ei", "<f4"),
        ("Ei_vs", "|u1"),
        ("ti", "<f4"),
        ("Imax", "<f4"),
        ("Imax_unit", "|u1"),
        ("Imin", "<f4"),
        ("Imin_unit", "|u1"),
        ("dQM", "<f4"),
        ("dQM_unit", "|u1"),
        ("record", "|u1"),
        ("dI", "<f4"),
        ("dI_unit", "|u1"),
        ("dQ", "<f4"),
        ("dQ_unit", "|u1"),
        ("dt", "<f4"),
        ("dta", "<f4"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("I_range", "|u1"),
        ("I_range_min", "|u1"),
        ("I_range_max", "|u1"),
        ("I_range_init", "|u1"),
        ("bandwidth", "u1"),
        ("goto_Ns", "<u4"),
        ("nc_cycles", "<u4"),
    ]
)


########################## Chronopotentiometry #########################
def _cp_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the CP technique."""
    params = [
        "Is",
        "Is_unit",
        "Is_vs",
        "ts",
        "EM",
        "dQM",
        "dQM_unit",
        "record",
        "dEs",
        "dts",
        "E_range_min",
        "E_range_max",
        "I_range",
        "bandwidth",
        "goto_Ns",
        "nc_cycles",
    ]
    return _prepend_ns(settings, params)


_cp_params_dtype = np.dtype(
    [
        ("Is", "<f4"),
        ("Is_unit", "|u1"),
        ("Is_vs", "|u1"),
        ("ts", "<f4"),
        ("EM", "<f4"),
        ("dQM", "<f4"),
        ("dQM_unit", "|u1"),
        ("record", "|u1"),
        ("dEs", "<f4"),
        ("dts", "<f4"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("I_range", "|u1"),
        ("bandwidth", "|u1"),
        ("goto_Ns", "<u4"),
        ("nc_cycles", "<u4"),
    ]
)

########################## Cyclic Voltammetry ##########################
def _cv_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the CV technique."""
    params = [
        "Ei",
        "Ei_vs",
        "dE/dt",
        "dE/dt_unit",
        "E1",
        "E1_vs",
        "step_percent",
        "N",
        "E_range_min",
        "E_range_max",
        "I_range",
        "I_range_min",
        "I_range_max",
        "I_range_init",
        "bandwidth",
        "E2",
        "E2_vs",
        "nc_cycles",
        "reverse_scan",
        "Ef",
        "Ef_vs",
    ]
    return params


_cv_params_dtype = np.dtype(
    [
        ("Ei", "<f4"),
        ("Ei_vs", "|u1"),
        ("dE/dt", "<f4"),
        ("dE/dt_unit", "|u1"),
        ("E1", "<f4"),
        ("E1_vs", "|u1"),
        ("step_percent", "|u1"),
        ("N", "<u4"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("I_range", "|u1"),
        ("I_range_min", "|u1"),
        ("I_range_max", "|u1"),
        ("I_range_init", "|u1"),
        ("bandwidth", "u1"),
        ("E2", "<f4"),
        ("E2_vs", "|u1"),
        ("nc_cycles", "<u4"),
        ("reverse_scan", "|u1"),
        ("Ef", "<f4"),
        ("Ef_vs", "|u1"),
    ]
)

############ Galvanostatic Cycling with Potential Limitation ###########
def _gcpl_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the GCPL technique."""
    params = [
        "set_I/C",
        "Is",
        "Is_unit",
        "Is_vs",
        "N",
        "I_sign",
        "t1",
        "I_range",
        "bandwidth",
        "dE1",
        "dt1",
        "EM",
        "tM",
        "Im",
        "Im_unit",
        "dI/dt",
        "dI/dt_unit",
        "E_range_min",
        "E_range_max",
        "dq",
        "dq_unit",
        "dtq",
        "dQM",
        "dQM_unit",
        "dxM",
        "tR",
        "dER/dt",
        "dER",
        "dtR",
        "EL",
        "goto_Ns",
        "nc_cycles",
    ]
    return _prepend_ns(settings, params)


_gcpl_params_dtype = np.dtype(
    [
        ("set_I/C", "|u1"),
        ("Is", "<f4"),
        ("Is_unit", "|u1"),
        ("Is_vs", "|u1"),
        ("N", "f4"),
        ("I_sign", "|u1"),
        ("t1", "<f4"),
        ("I_range", "|u1"),
        ("bandwidth", "|u1"),
        ("dE1", "<f4"),
        ("dt1", "<f4"),
        ("EM", "<f4"),
        ("tM", "<f4"),
        ("Im", "<f4"),
        ("Im_unit", "|u1"),
        ("dI/dt", "<f4"),
        ("dI/dt_unit", "|u1"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("dq", "<f4"),
        ("dq_unit", "|u1"),
        ("dtq", "<f4"),
        ("dQM", "<f4"),
        ("dQM_unit", "|u1"),
        ("dxM", "<f4"),
        ("tR", "<f4"),
        ("dER/dt", "<f4"),
        ("dER", "<f4"),
        ("dtR", "<f4"),
        ("EL", "<f4"),
        ("goto_Ns", "<u4"),
        ("nc_cycles", "<u4"),
    ]
)

############ Galvano Electrochemical Impedance Spectroscopy ############
def _geis_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the GEIS technique."""
    params = [
        "sine_mode",
        "Is",
        "Is_unit",
        "Is_vs",
        "tIs",
        "record",
        "dE",
        "dt",
        "fi",
        "fi_unit",
        "ff",
        "ff_unit",
        "Nd",
        "points",
        "spacing",
        "Ia/Va",
        "Ia",
        "Ia_unit",
        "va_pourcent",  # Random French parameter name?
        "pw",
        "Na",
        "corr",
    ]
    params = _prepend_ns(settings, params)
    lim_nb_match = re.search(r"^lim_nb\s+(?P<val>.+)", "\n".join(settings))
    if lim_nb_match:
        lim_nb = int(max(lim_nb_match["val"].split()))
        params.append("lim_nb")
        for i, __ in enumerate(range(lim_nb), 1):
            limit = [
                f"limit_type{i}",
                f"limit_comp{i}",
                f"limit_value{i}",
                f"limit_unit{i}",
            ]
            params += limit
    params += [
        "E_range_min",
        "E_range_max",
        "I_range",
        "bandwidth",
        "nc_cycles",
        "goto_Ns",
        "nr_cycles",
        "inc_cycle",
    ]
    return params


_geis_params_dtype = np.dtype(
    [
        ("sine_mode", "|u1"),
        ("Is", "<f4"),
        ("Is_unit", "|u1"),
        ("Is_vs", "|u1"),
        ("tIs", "<f4"),
        ("record", "|u1"),
        ("dE", "<f4"),
        ("dt", "<f4"),
        ("fi", "<f4"),
        ("fi_unit", "|u1"),
        ("ff", "<f4"),
        ("ff_unit", "|u1"),
        ("Nd", "<u4"),
        ("points", "|u1"),
        ("spacing", "|u1"),
        ("Ia/Va", "|u1"),
        ("Ia", "<f4"),
        ("Ia_unit", "|u1"),
        ("va_pourcent", "<f4"),
        ("pw", "<f4"),
        ("Na", "<u4"),
        ("corr", "|u1"),
        ("lim_nb", "|u1"),
        ("limit_type1", "|u1"),
        ("limit_comp1", "|u1"),
        ("limit_value1", "<f4"),
        ("limit_unit1", "|u1"),
        ("limit_type2", "|u1"),
        ("limit_comp2", "|u1"),
        ("limit_value2", "<f4"),
        ("limit_unit2", "|u1"),
        ("limit_type3", "|u1"),
        ("limit_comp3", "|u1"),
        ("limit_value3", "<f4"),
        ("limit_unit3", "|u1"),
        ("limit_type4", "|u1"),
        ("limit_comp4", "|u1"),
        ("limit_value4", "<f4"),
        ("limit_unit4", "|u1"),
        ("limit_type5", "|u1"),
        ("limit_comp5", "|u1"),
        ("limit_value5", "<f4"),
        ("limit_unit5", "|u1"),
        ("limit_type6", "|u1"),
        ("limit_comp6", "|u1"),
        ("limit_value6", "<f4"),
        ("limit_unit6", "|u1"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("I_range", "|u1"),
        ("bandwidth", "|u1"),
        ("nc_cycles", "<u4"),
        ("goto_Ns", "<u4"),
        ("nr_cycles", "<u4"),
        ("inc_cycle", "<u4"),
    ]
)

################################# Loop #################################
def _loop_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the LOOP technique."""
    params = ["goto_Ne", "nt_times"]
    return params


####################### Linear Sweep Voltammetry #######################
def _lsv_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the LSV technique."""
    params = [
        "tR",
        "dER/dt",
        "dER",
        "dtR",
        "dE/dt",
        "dE/dt_unit",
        "Ei",
        "Ei_vs",
        "EL",
        "EL_vs",
        "record",
        "dI",
        "dI_unit",
        "tI",
        "step_percent",
        "N",
        "E_range_min",
        "E_range_max",
        "I_range",
        "I_range_min",
        "I_range_max",
        "I_range_init",
        "bandwidth",
    ]
    return params


####################### Linear Sweep Voltammetry #######################
_lsv_params_dtype = np.dtype(
    [
        ("tR", "<f4"),
        ("dER/dt", "<f4"),
        ("dER", "<f4"),
        ("dtR", "<f4"),
        ("dE/dt", "<f4"),
        ("dE/dt_unit", "|u1"),
        ("Ei", "<f4"),
        ("Ei_vs", "|u1"),
        ("EL", "<f4"),
        ("EL_vs", "|u1"),
        ("record", "|u1"),
        ("dI", "<f4"),
        ("dI_unit", "|u1"),
        ("tI", "<f4"),
        ("step_percent", "|u1"),
        ("N", "<u4"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("I_range", "|u1"),
        ("I_range_min", "|u1"),
        ("I_range_max", "|u1"),
        ("I_range_init", "|u1"),
        ("bandwidth", "|u1"),
    ]
)


############################## Modulo Bat ##############################
def _mb_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the MB technique."""
    params = [
        "ctrl_type",
        "apply_I/C",
        "ctrl1_val",
        "ctrl1_val_unit",
        "ctrl1_val_vs",
        "ctrl2_val",
        "ctrl2_val_unit",
        "ctrl2_val_vs",
        "ctrl3_val",
        "ctrl3_val_unit",
        "ctrl3_val_vs",
        "N",
        "charge/discharge",
    ]
    params = _prepend_ns(settings, params)
    n1_match = re.search(r"^N1\s+", "\n".join(settings))
    if n1_match:
        n1 = [
            "charge/discharge_1",
            "apply_I/C_1",
            "N1",
            "ctrl4_val_unit",
            "ctrl4_val_vs",
        ]
        params += n1
    params += [
        "ctrl_seq",
        "ctrl_repeat",
        "ctrl_trigger",
        "ctrl_TO_t",
        "ctrl_TO_t_unit",
        "ctrl_Nd",
        "ctrl_Na",
        "ctrl_corr",
    ]
    lim_nb_match = re.search(r"^lim_nb\s+(?P<val>.+)", "\n".join(settings))
    if lim_nb_match:
        lim_nb = int(max(lim_nb_match["val"].split()))
        params.append("lim_nb")
        for i, __ in enumerate(range(lim_nb), 1):
            lim = [
                f"lim{i}_type",
                f"lim{i}_comp",
                f"lim{i}_Q",
                f"lim{i}_value",
                f"lim{i}_value_unit",
                f"lim{i}_action",
                f"lim{i}_seq",
            ]
            params += lim
    rec_nb_match = re.search(r"^rec_nb\s+(?P<val>.+)", "\n".join(settings))
    if rec_nb_match:
        rec_nb = int(max(rec_nb_match["val"].split()))
        params.append("rec_nb")
        for i, __ in enumerate(range(rec_nb), 1):
            rec = [f"rec{i}_type", f"rec{i}_value", f"rec{i}_value_unit"]
            params += rec
    params += [
        "E_range_min",
        "E_range_max",
        "I_range",
        "I_range_min",
        "I_range_max",
        "I_range_init",
        "auto_rest",
        "bandwidth",
    ]
    return params


_mb_params_dtypes = [
    np.dtype(
        [
            ("ctrl_type", "|u1"),
            ("apply_I/C", "|u1"),
            ("ctrl1_val", "<f4"),
            ("ctrl1_val_unit", "|u1"),
            ("ctrl1_val_vs", "|u1"),
            ("ctrl2_val", "<f4"),
            ("ctrl2_val_unit", "|u1"),
            ("ctrl2_val_vs", "|u1"),
            ("ctrl3_val", "<f4"),
            ("ctrl3_val_unit", "|u1"),
            ("ctrl3_val_vs", "|u1"),
            ("N", "<f4"),
            ("charge/discharge", "|u1"),
            ("charge/discharge_1", "|u1"),
            ("apply_I/C_1", "|u1"),
            ("N1", "<f4"),
            ("ctrl4_val", "<f4"),
            ("ctrl4_val_unit", "|u1"),
            ("ctrl_seq", "<u4"),
            ("ctrl_repeat", "<u4"),
            ("ctrl_trigger", "|u1"),
            ("ctrl_TO_t", "<f4"),
            ("ctrl_TO_t_unit", "|u1"),
            ("ctrl_Nd", "<u4"),
            ("ctrl_Na", "<u4"),
            ("ctrl_corr", "|u1"),
            ("lim_nb", "|u1"),
            ("lim1_type", "|u1"),
            ("lim1_comp", "|u1"),
            ("lim1_Q", "|u1"),
            ("lim1_value", "<f4"),
            ("lim1_value_unit", "|u1"),
            ("lim1_action", "|u1"),
            ("lim1_seq", "<u4"),
            ("lim2_type", "|u1"),
            ("lim2_comp", "|u1"),
            ("lim2_Q", "|u1"),
            ("lim2_value", "<f4"),
            ("lim2_value_unit", "|u1"),
            ("lim2_action", "|u1"),
            ("lim2_seq", "<u4"),
            ("lim3_type", "|u1"),
            ("lim3_comp", "|u1"),
            ("lim3_Q", "|u1"),
            ("lim3_value", "<f4"),
            ("lim3_value_unit", "|u1"),
            ("lim3_action", "|u1"),
            ("lim3_seq", "<u4"),
            ("rec_nb", "|u1"),
            ("rec1_type", "|u1"),
            ("rec1_value", "<f4"),
            ("rec1_value_unit", "|u1"),
            ("rec2_type", "|u1"),
            ("rec2_value", "<f4"),
            ("rec2_value_unit", "|u1"),
            ("rec3_type", "|u1"),
            ("rec3_value", "<f4"),
            ("rec3_value_unit", "|u1"),
            ("E_range_min", "<f4"),
            ("E_range_max", "<f4"),
            ("I_range", "|u1"),
            ("I_range_min", "|u1"),
            ("I_range_max", "|u1"),
            ("I_range_init", "|u1"),
            ("auto_rest", "|u1"),
            ("bandwidth", "|u1"),
        ]
    ),
    np.dtype(
        [
            ("ctrl_type", "|u1"),
            ("apply_I/C", "|u1"),
            ("ctrl1_val", "<f4"),
            ("ctrl1_val_unit", "|u1"),
            ("ctrl1_val_vs", "|u1"),
            ("ctrl2_val", "<f4"),
            ("ctrl2_val_unit", "|u1"),
            ("ctrl2_val_vs", "|u1"),
            ("ctrl3_val", "<f4"),
            ("ctrl3_val_unit", "|u1"),
            ("ctrl3_val_vs", "|u1"),
            ("N", "<f4"),
            ("charge/discharge", "|u1"),
            ("ctrl_seq", "<u4"),
            ("ctrl_repeat", "<u4"),
            ("ctrl_trigger", "|u1"),
            ("ctrl_TO_t", "<f4"),
            ("ctrl_TO_t_unit", "|u1"),
            ("ctrl_Nd", "<u4"),
            ("ctrl_Na", "<u4"),
            ("ctrl_corr", "|u1"),
            ("lim_nb", "|u1"),
            ("lim1_type", "|u1"),
            ("lim1_comp", "|u1"),
            ("lim1_Q", "|u1"),
            ("lim1_value", "<f4"),
            ("lim1_value_unit", "|u1"),
            ("lim1_action", "|u1"),
            ("lim1_seq", "<u4"),
            ("lim2_type", "|u1"),
            ("lim2_comp", "|u1"),
            ("lim2_Q", "|u1"),
            ("lim2_value", "<f4"),
            ("lim2_value_unit", "|u1"),
            ("lim2_action", "|u1"),
            ("lim2_seq", "<u4"),
            ("lim3_type", "|u1"),
            ("lim3_comp", "|u1"),
            ("lim3_Q", "|u1"),
            ("lim3_value", "<f4"),
            ("lim3_value_unit", "|u1"),
            ("lim3_action", "|u1"),
            ("lim3_seq", "<u4"),
            ("rec_nb", "|u1"),
            ("rec1_type", "|u1"),
            ("rec1_value", "<f4"),
            ("rec1_value_unit", "|u1"),
            ("rec2_type", "|u1"),
            ("rec2_value", "<f4"),
            ("rec2_value_unit", "|u1"),
            ("rec3_type", "|u1"),
            ("rec3_value", "<f4"),
            ("rec3_value_unit", "|u1"),
            ("E_range_min", "<f4"),
            ("E_range_max", "<f4"),
            ("I_range", "|u1"),
            ("I_range_min", "|u1"),
            ("I_range_max", "|u1"),
            ("I_range_init", "|u1"),
            ("auto_rest", "|u1"),
            ("bandwidth", "|u1"),
        ]
    ),
]

######################### Open Circuit Voltage #########################
def _ocv_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the OCV technique."""
    params = ["tR", "dER/dt"]
    record_match = re.search(r"^record.+", "\n".join(settings))
    if record_match:
        params += ["record"]
    params += ["dER", "dtR", "E_range_min", "E_range_max"]
    return params


_ocv_params_dtypes = [
    np.dtype(
        [
            ("tR", "<f4"),
            ("dER/dt", "<f4"),
            ("record", "|u1"),
            ("dER", "<f4"),
            ("dtR", "<f4"),
            ("E_range_min", "<f4"),
            ("E_range_max", "<f4"),
        ]
    ),
    np.dtype(
        [
            ("tR", "<f4"),
            ("dER/dt", "<f4"),
            ("dER", "<f4"),
            ("dtR", "<f4"),
            ("E_range_min", "<f4"),
            ("E_range_max", "<f4"),
        ]
    ),
]


########### Potentio Electrochemical Impedance Spectroscopy ############
def _peis_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the PEIS technique."""
    params = [
        "sine_mode",
        "E",
        "E_vs",
        "tE",
        "record",
        "dI",
        "dI_unit",
        "dt",
        "fi",
        "fi_unit",
        "ff",
        "ff_unit",
        "Nd",
        "points",
        "spacing",
        "Va",
        "pw",
        "Na",
        "corr",
    ]
    params = _prepend_ns(settings, params)
    lim_nb_match = re.search(r"^lim_nb\s+(?P<val>.+)", "\n".join(settings))
    if lim_nb_match:
        lim_nb = int(max(lim_nb_match["val"].split()))
        params.append("lim_nb")
        for i, __ in enumerate(range(lim_nb), 1):
            limit = [
                f"limit_type{i}",
                f"limit_comp{i}",
                f"limit_value{i}",
                f"limit_unit{i}",
            ]
            params += limit
    params += [
        "E_range_min",
        "E_range_max",
        "I_range",
        "bandwidth",
        "nc_cycles",
        "goto_Ns",
        "nr_cycles",
        "inc_cycle",
    ]
    return params


_peis_params_dtype = np.dtype(
    [
        ("sine_mode", "|u1"),
        ("E", "<f4"),
        ("E_vs", "|u1"),
        ("tE", "<f4"),
        ("record", "|u1"),
        ("dI", "<f4"),
        ("dI_unit", "|u1"),
        ("dt", "<f4"),
        ("fi", "<f4"),
        ("fi_unit", "|u1"),
        ("ff", "<f4"),
        ("ff_unit", "|u1"),
        ("Nd", "<u4"),
        ("points", "|u1"),
        ("spacing", "|u1"),
        ("Va", "<f4"),
        ("pw", "<f4"),
        ("Na", "<u4"),
        ("corr", "|u1"),
        ("lim_nb", "|u1"),
        ("limit_type1", "|u1"),
        ("limit_comp1", "|u1"),
        ("limit_value1", "<f4"),
        ("limit_unit1", "|u1"),
        ("limit_type2", "|u1"),
        ("limit_comp2", "|u1"),
        ("limit_value2", "<f4"),
        ("limit_unit2", "|u1"),
        ("limit_type3", "|u1"),
        ("limit_comp3", "|u1"),
        ("limit_value3", "<f4"),
        ("limit_unit3", "|u1"),
        ("limit_type4", "|u1"),
        ("limit_comp4", "|u1"),
        ("limit_value4", "<f4"),
        ("limit_unit4", "|u1"),
        ("limit_type5", "|u1"),
        ("limit_comp5", "|u1"),
        ("limit_value5", "<f4"),
        ("limit_unit5", "|u1"),
        ("limit_type6", "|u1"),
        ("limit_comp6", "|u1"),
        ("limit_value6", "<f4"),
        ("limit_unit6", "|u1"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("I_range", "|u1"),
        ("bandwidth", "|u1"),
        ("nc_cycles", "<u4"),
        ("goto_Ns", "<u4"),
        ("nr_cycles", "<u4"),
        ("inc_cycle", "<u4"),
    ]
)


################################# Wait #################################
def _wait_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the WAIT technique."""
    params = [
        "select",
        "td",
        "from",
        "tech_num",
        "ole_date",
        "ole_time",
        "record",
        "dE",
        "dI",
        "dI_unit",
        "dt",
    ]
    return params


_wait_params_dtype = np.dtype(
    [
        ("select", "|u1"),
        ("td", "<u4"),
        ("from", "|u1"),
        ("tech_num", "|u1"),
        ("ole_date", "<f4"),  # Why the hell would they split this?!
        ("ole_time", "<f4"),
        ("record", "|u1"),
        ("dE", "<f4"),
        ("dI", "<f4"),
        ("dI_unit", "|u1"),
        ("dt", "<f4"),
    ]
)


######################## IR compensation (PEIS) ########################
def _zir_params(settings: list[str]) -> list[str]:
    """Constructs the parameter names for the ZIR technique."""
    params = [
        "E",
        "E_vs",
        "f",
        "f_unit",
        "Va",
        "pw",
        "Na",
        "E_range_min",
        "E_range_max",
        "I_range",
        "bandwidth",
        "comp_level",
        "use_results",
        "comp_mode",
    ]
    return params


_zir_params_dtype = np.dtype(
    [
        ("E", "<f4"),
        ("E_vs", "|u1"),
        ("f", "<f4"),
        ("f_unit", "|u1"),
        ("Va", "<f4"),
        ("pw", "<f4"),
        ("Na", "<u4"),
        ("E_range_min", "<f4"),
        ("E_range_max", "<f4"),
        ("I_range", "|u1"),
        ("bandwidth", "|u1"),
        ("comp_level", "|u1"),
        ("use_results", "|u1"),
        ("comp_mode", "|u1"),
    ]
)


def technique_params(technique: str, settings: list[str]) -> tuple[str, list]:
    """Constructs the parameter names for different techniques.

    Parameters
    ----------
    technique
        The full name of the technique.

    settings
        The list of settings from the start of an .mpt or .mps file.

    Returns
    -------
    tuple[str, list]
        The short technique name and a full list of technique parameter
        names depending on what is present in the file.

    """
    if technique == "Chronoamperometry / Chronocoulometry":
        return "CA", _ca_params(settings)
    if technique == "Chronopotentiometry":
        return "CP", _cp_params(settings)
    if technique == "Cyclic Voltammetry":
        return "CV", _cv_params(settings)
    if technique == "Galvano Electrochemical Impedance Spectroscopy":
        return "GEIS", _geis_params(settings)
    if technique == "Galvanostatic Cycling with Potential Limitation":
        return "GCPL", _gcpl_params(settings)
    if technique == "IR compensation (PEIS)":
        return "ZIR", _zir_params(settings)
    if technique == "Linear Sweep Voltammetry":
        return "LSV", _lsv_params(settings)
    if technique == "Loop":
        return "LOOP", _loop_params(settings)
    if technique == "Modulo Bat":
        return "mb", _mb_params(settings)
    if technique == "Open Circuit Voltage":
        return "OCV", _ocv_params(settings)
    if technique == "Potentio Electrochemical Impedance Spectroscopy":
        return "PEIS", _peis_params(settings)
    if technique == "Wait":
        return "WAIT", _wait_params(settings)
    raise NotImplementedError(f"'{technique}' not implemented.")


# Maps the technique byte to its corresponding dtype.
technique_params_dtypes = {
    0x04: ("GCPL", _gcpl_params_dtype),
    0x06: ("CV", _cv_params_dtype),
    0x0B: ("OCV", _ocv_params_dtypes),
    0x18: ("CA", _ca_params_dtype),
    0x19: ("CP", _cp_params_dtype),
    0x1C: ("WAIT", _wait_params_dtype),
    0x1D: ("PEIS", _peis_params_dtype),
    0x1E: ("GEIS", _geis_params_dtype),
    0x32: ("ZIR", _zir_params_dtype),
    0x6C: ("LSV", _lsv_params_dtype),
    0x7F: ("MB", _mb_params_dtypes),
}

param_map = {
    "I_range": (
        ("1 A", 9, 1),
        ("100 mA", 10, 1e-1),
        ("10 mA", 11, 1e-2),
        ("1 mA", 12, 1e-3),
        ("100 µA", 13, 1e-4),
        ("10 µA", 14, 1e-5),
        ("1 µA", 15, 1e-6),
        ("Auto", 21, None),
        ("Auto", 23, None),  # guess
        ("Auto", 24, None),  # guess
        ("1 A", 37, 1),
    ),
    "Is_unit": (
        ("A", 0),  # guess
        ("mA", 1),
        ("µA", 2),
        ("nA", 3),  # guess
        ("pA", 4),  # guess
    ),
}


def param_from_key(
    param: str, key: Union[int, str], to_str: bool = True
) -> Union[str, float]:
    """
    Convert a supplied key of a certain parameter to its string or float value.

    The function uses the map defined in ``param_map`` to convert between the
    entries in the tuples, which contain the :class:`str` value of the parameter
    (present in ``.mpt`` files), the :class:`int` value of the parameter (present
    in ``.mpr`` files), and the corresponding :class:`float` value in SI units.

    Parameters
    ----------
    param
        The name of the parameter, a key within the ``param_map``. If ``param``
        is not present in ``param_map``, the supplied key is returned back.

    key
        The key of the parameter that is to be converted to a different representation.

    to_str
        A switch between :class:`str` and :class:`float` output.

    Returns
    -------
    key: Union[str, float, int]
        The key converted to the requested format.

    """
    ii = 1 if isinstance(key, int) else 0
    if param in param_map:
        for i in param_map[param]:
            if i[ii] == key:
                if to_str:
                    return i[0]
                else:
                    return i[2]
        raise ValueError(f"element '{key}' for parameter '{param}' not understood.")
    return key


def get_resolution(name: str, value: float, Erange: float, Irange: float) -> float:
    """
    Function that returns the resolution of a property based on its name, value,
    E-range and I-range.

    The values used here are hard-coded from VMP-3 potentiostats. Generally, the
    resolution is returned, however in some cases only the accuracy is specified
    (currently ``freq`` and ``Phase``).

    """
    if name in ["control_V"]:
        # VMP-3: bisect function between 5 µV and 300 µV, as the
        # voltage is stored in a 16-bit int.
        if Erange >= 20.0:
            return 305.18e-6
        else:
            res = [5e-6, 10e-6, 20e-6, 50e-6, 100e-6, 150e-6, 200e-6, 300e-6, 305.18e-6]
            i = bisect.bisect_right(res, Erange / np.iinfo(np.uint16).max)
            return res[i]
    elif name in ["Ewe", "Ece", "|Ewe|", "|Ece|"]:
        # VMP-3: 0.0015% of FSR, 75 µV minimum
        return max(Erange * 0.0015 / 100, 75e-6)
    elif name in ["control_I"]:
        # VMP-3: 0.004% of FSR, 760 µV at 10 µA I-range
        return max(Irange * 0.004 / 100, 760e-12)
    elif name in ["I", "|I|"]:
        # VMP-3: 0.004% of FSR
        if Irange is None:
            logging.warning("get_resolution: 'Irange' not specified. Using 'I'.")
            Irange = 10 ** math.ceil(math.log10(value))
        return Irange * 0.004 / 100
    elif name in ["freq"]:
        # VMP-3: using accuracy: 1% of value
        return value * 0.01
    elif name in ["Phase(Z)", "Phase(Y)"]:
        # VMP-3: using accuracy: 1 degree
        return 1.0
    elif name in ["|Z|", "Re(Z)", "-Im(Z)"]:
        # |Z| = |Ewe|/|I|; assuming GEIS
        return get_resolution("I", value, Erange, Irange)
    else:
        return math.ulp(value)



from yadg.parsers.masstrace import quadstarsac


version = "4.0.0"


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    tracetype: str = "quadstar.sac",
) -> tuple[list, dict, bool]:
    """Unified mass spectrometry data parser.

    This parser processes mass spectrometry scans in signal(mass) format.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    tracetype
        Determines the output file format. Currently supported formats
        can be found :ref:`here<parsers_masstrace_formats>`.

        The default is ``"quadstar.sac"``.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag.

    """
    if tracetype == "quadstar.sac":
        _data, _meta = quadstarsac.process(fn, encoding, timezone)
    return _data, _meta, True



import json
import numpy as np
import logging
import uncertainties as uc
import uncertainties.unumpy as unp

from yadg.parsers.chromtrace import (
    ezchromasc,
    agilentcsv,
    agilentch,
    agilentdx,
    fusionjson,
    fusionzip,
)
from yadg.parsers.chromtrace import integration

version = "4.1.0"


def parse_detector_spec(
    calfile: str = None, detectors: dict = None, species: dict = None
) -> dict:
    """
    Chromatography detector parser.

    Combines the specification provided in ``calfile`` with that provided in
    ``detectors`` and ``species``.

    .. _parsers_chromtrace_calfile:

    The format of ``calfile`` is as follows:

    .. code-block:: yaml

        "{{ detector_name }}":    # name of the detector
          id:           !!int     # ID of the detector used for matching
          prefer:       !!bool    # whether to prefer this detector for xout calc
          peakdetect:
            window:     !!int     # Savigny-Golay window_length = 2*window + 1
            polyorder:  !!int     # Savigny-Golay polyorder
            prominence: !!float   # peak picking prominence parameter
            threshold:  !!float   # peak edge detection threshold
          species:
            "{{ species_name }}": # name of the analyte
              l:        !!float   # peak picking left limit [s]
              r:        !!float   # peak picking right limit [s]
              calib:    {}        # calibration specification
              unit:     !!str     # optional unit for the concentration, by default %

    .. note::
        The syntax of the calibration specification is detailed in
        :func:`yadg.dgutils.calib.calib_handler`.

    .. _parsers_chromtrace_detectors:

    The format of ``detectors`` is as follows:

    .. code-block:: yaml

        "{{ detector_name }}":  # name of the detector
          id:           !!int   # ID of the detector used for matching
          prefer:       !!bool  # whether to prefer this detector for xout calc
          peakdetect:
            window:     !!int   # Savigny-Golay window_length = 2*window + 1
            polyorder:  !!int   # Savigny-Golay polyorder
            prominence: !!float # peak picking prominence parameter
            threshold:  !!float # peak edge detection threshold

    .. _parsers_chromtrace_species:

    The format of ``species`` is as follows:

    .. code-block:: yaml

        "{{ detector_name }}":    # name of the detector
          species:
            "{{ species_name }}": # name of the analyte
              l:        !!float   # peak picking left limit [s]
              r:        !!float   # peak picking right limit [s]
              calib:    !!calib   # calibration specification

    .. note::
        The syntax of the calibration specification is detailed in
        :func:`yadg.dgutils.calib.calib_handler`.

    Parameters
    ----------
    calfile
        A json file containing the calibration data in the format prescribed
        :ref:`above<parsers_chromtrace_calfile>`.

    detectors
        A dictionary containing the ``"id"``, ``"peakdetect"`` and ``"prefer"``
        keys for each detector, as shown :ref:`here<parsers_chromtrace_detectors>`.

    species
        A dictionary containing the species names as keys and their specification
        as dictionaries, as shown :ref:`here<parsers_chromtrace_species>`.

    Returns
    -------
    calib: dict
        The combined calibration specification.
    """

    if calfile is not None:
        with open(calfile, "r") as infile:
            calib = json.load(infile)
    else:
        calib = {}
    if isinstance(detectors, dict):
        for name, det in detectors.items():
            try:
                calib[name].update(det)
            except KeyError:
                calib[name] = det
    if isinstance(species, dict):
        for name, sp in species.items():
            assert name in calib, (
                f"chromtrace: Detector with name {name} specified in supplied "
                f"'species' but previously undefined."
            )
            try:
                calib[name]["species"].update(sp)
            except KeyError:
                calib[name]["species"] = sp
    return calib


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    tracetype: str = "ezchrom.asc",
    detectors: dict = None,
    species: dict = None,
    calfile: str = None,
) -> tuple[list, dict, bool]:
    """
    Unified chromatogram parser.

    This parser processes GC and LC chromatograms in signal(time) format. When
    provided with a calibration file, this tool will integrate the trace, and provide
    the peak areas, retention times, and concentrations of the detected species.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    tracetype
        Determines the output file format. Currently supported formats are:

        -  ``"ezchrom.asc"`` (EZ-Chrom ASCII export),
        -  ``"agilent.csv"`` (Agilent Chemstation chromtab csv format),
        -  ``"agilent.ch"`` (Agilent OpenLab binary signal file),
        -  ``"agilent.dx"`` (Agilent OpenLab binary data archive),
        -  ``"fusion.json"`` (Fusion json file),
        -  ``"fusion.zip"`` (Fusion zip file),

        The default is ``"ezchrom.asc"``.

    detectors
        Detector specification. Matches and identifies a trace in the `fn` file.
        If provided, overrides data provided in ``calfile``, below.

    species
        Species specification. Per-detector species can be listed here, providing an
        expected retention time range for the peak maximum. Additionally, calibration
        data can be supplied here. Overrides data provided in ``calfile``, below.

    calfile
        Path to a json file containing the ``detectors`` and ``species`` spec. Either
        ``calfile`` and/or ``species`` and ``detectors`` have to be provided.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. All currently
        supported file formats return full date.
    """
    if tracetype == "ezchrom.asc":
        _data, _meta = ezchromasc.process(fn, encoding, timezone)
    elif tracetype == "agilent.csv":
        _data, _meta = agilentcsv.process(fn, encoding, timezone)
    elif tracetype == "agilent.dx":
        _data, _meta = agilentdx.process(fn, encoding, timezone)
    elif tracetype == "agilent.ch":
        _data, _meta = agilentch.process(fn, encoding, timezone)
    elif tracetype == "fusion.json":
        _data, _meta = fusionjson.process(fn, encoding, timezone)
    elif tracetype == "fusion.zip":
        _data, _meta = fusionzip.process(fn, encoding, timezone)

    if calfile is None and (species is None or detectors is None):
        logging.warning(
            "chromtrace: Neither 'calfile' nor both 'species' and 'detectors' were "
            "provided. Will proceed without peak integration."
        )
        chromspec = False
    else:
        chromspec = parse_detector_spec(calfile, detectors, species)

    results = []
    for chrom in _data:
        result = {}
        # process derived data
        if chromspec:
            result["derived"] = integration.integrate_trace(chrom["traces"], chromspec)
        else:
            for k, v in chrom["traces"].items():
                v.pop("data")
        # process raw data here
        result["uts"] = chrom.pop("uts")
        result["fn"] = chrom.pop("fn")
        result["raw"] = chrom

        results.append(result)

    return results, _meta, True



from yadg.parsers.electrochem import eclabmpr, eclabmpt

version = "4.1.0"


def process(
    fn: str,
    encoding: str = "windows-1252",
    timezone: str = "localtime",
    filetype: str = "eclab.mpr",
) -> tuple[list, dict, bool]:
    """Unified parser for electrochemistry data.

    Parameters
    ----------
    fn
        The file containing the data to parse.

    encoding
        Encoding of ``fn``, by default "windows-1252".

    timezone
        A string description of the timezone. Default is "localtime".

    filetype
        Determines the output file format. Currently supported formats
        can be found :ref:`here<parsers_electrochem_formats>`.

        The default is ``eclab.mpr`` (auto).

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. The currently
        implemented parsers all return full date.

    """
    if filetype == "eclab.mpr":
        data, meta, fulldate = eclabmpr.process(fn, encoding, timezone)
    elif filetype == "eclab.mpt":
        data, meta, fulldate = eclabmpt.process(fn, encoding, timezone)
    return data, meta, fulldate



import sys
from importlib import metadata as ilmd


def get_yadg_metadata() -> dict:
    """
    Returns current **yadg** metadata.
    """
    metadata = {"version": ilmd.version("yadg"), "command": " ".join(sys.argv)}
    return metadata



"""
.. _parsers_chromtrace_integration:

This module contains the :func:`yadg.parsers.chromtrace.integration.integrate_trace`
function, as well as several helper functions to smoothen, peak-pick, determine edges,
and integrate the supplied traces.

Smoothing
`````````
Smoothing can be optionally performed on the Y-values of each trace, using a 
Savigny-Golay filter. The default smoothing is performed using a cubic fit to a window 
length of 7; if the polyorder or the window length are not specified, smoothing is 
not used.

Peak-picking and edge-finding
`````````````````````````````
Peak-picking is performed on the smoothed Y-data to find peaks, as well as on the mirror
image of the data to find bands. Only peaks are further processed. Additionally, the
1st and 2nd derivatives of the Y-data are evaluated, and the zero-points are found using
numpy routines.

The peak edges are taken as either the nearest minima adjacent to the peak maximum,
or as the inflection points at which the gradient falls below a prescribed threshold,
whichever is closest to the peak maximum.

Baseline correction
```````````````````
Using the determined peak-edges, the baseline is linearly interpolated in sections of
Y-data which belong to a peak. The interpolation is performed using the raw (not 
smoothened) Y-data. 

If multiple peaks are adjacent to each other without a gap, the interpolation begins
at the left limit of the leftmost peak and continues uninterrupted to the right limit
of the rightmost peak. The points which belong to the interpolated areas are assumed 
to have an uncertainty of zero.

The corrected baseline is then obtained by subtracting the interpolated baseline from
the original raw (not smoothened) data.

Peak integration
````````````````
Peak integration is performed on the corrected baseline and the matching X-data using
the trapezoidal method as implemented in ``np.trapz``.

.. codeauthor:: Peter Kraus <peter.kraus@empa.ch>
"""
import numpy as np
from scipy.signal import savgol_filter, find_peaks
import logging
import uncertainties as uc
from uncertainties import unumpy as unp

import yadg.dgutils


def _get_smooth_yvals(yvals: np.ndarray, pd: dict) -> np.ndarray:
    """
    Wrapper around :func:`scipy.signal.savgol_filter`. If the window length and the
    polynomial order are specified in ``pd``, the smoothened ``yvals`` are returned.
    Otherwise, the original ``yvals`` are returned.

    The function also validates that the specified window length derived from the
    ``"window"`` is larger than the ``"polyorder"``.
    """
    if pd.get("polyorder", None) is None or pd.get("window", None) is None:
        logging.info("chromtrace: no smoothing.")
        return yvals
    else:
        window = pd.get("window", 3) * 2 + 1
        polyorder = pd.get("polyorder", 3)
        assert polyorder < window, f"chromtrace: specified window <= polyorder."
        if polyorder == 2:
            logging.warning(
                "chromtrace: smoothing with a polyorder == 2 can be unreliable. "
                "Consider switching to a higher polyorder or disabling smoothing "
                "completely."
            )
        return savgol_filter(yvals, window_length=window, polyorder=polyorder)


def _find_peak_maxima(yvals: np.ndarray, pd: dict) -> dict:
    """
    This function is a wrapper around :func:`scipy.signal.find_peaks` as well as
    a helper finding the points at which the gradient and Hessian of ``yvals``
    is zero (maxima, minima, inflection points).
    """
    # find positive and negative peak indices
    peaks = {}
    res, _ = find_peaks(yvals, prominence=pd.get("prominence", 1e-4 * yvals.max()))
    peaks["+"] = res
    res, _ = find_peaks(yvals * -1, prominence=pd.get("prominence", 1e-4 * yvals.max()))
    peaks["-"] = res
    # gradient: find peaks and inflection points
    grad = np.gradient(yvals)
    res = np.nonzero(np.diff(np.sign(grad)))[0] + 1
    peaks["gradzero"] = res
    # second derivative: find peaks
    hess = np.gradient(grad)
    res = np.nonzero(np.diff(np.sign(hess)))[0] + 1
    peaks["hesszero"] = res
    return peaks, grad, hess


def _find_peak_edges(
    yvals: np.ndarray, ygrad: np.ndarray, peakdata: dict, detector: dict
) -> dict:
    """
    A function that, given the y-values of a trace in ``yvals`` and the maxima,
    inflection points etc. in ``peakdata``, and the peak integration `"threshold"`
    in ``detector``, finds the edges ``"llim"`` and ``"rlim"`` of each peak.
    """
    threshold = detector.get("threshold", 1.0)
    allpeaks = []
    for pi in range(len(peakdata["+"])):
        pmax = peakdata["+"][pi]
        for i in range(len(peakdata["hesszero"])):
            if peakdata["hesszero"][i] > pmax:
                hi = i
                break
        for i in range(len(peakdata["gradzero"])):
            if peakdata["gradzero"][i] > pmax:
                gi = i
                break
        # right of peak:
        rmin = False
        rthr = False
        for xi in range(peakdata["hesszero"][hi], yvals.size):
            if xi in peakdata["gradzero"] and not rmin:
                rmin = xi
            if xi in peakdata["hesszero"][hi:] and not rthr:
                if abs(ygrad[xi]) < threshold:
                    rthr = xi
            if rthr and rmin:
                break
        rlim = min(rthr if rthr else yvals.size, rmin if rmin else yvals.size)
        if rlim == yvals.size:
            logging.warning("gctrace: possible mismatch of peak end.")
            rlim -= 1
        # left of peak
        lmin = False
        lthr = False
        for xi in range(0, peakdata["hesszero"][hi - 1])[::-1]:
            if xi in peakdata["gradzero"] and not lmin:
                lmin = xi
            if xi in peakdata["hesszero"][: hi - 1] and not lthr:
                if abs(ygrad[xi]) < threshold:
                    lthr = xi
            if lthr and lmin:
                break
        llim = max(lthr if lthr else 0, lmin if lmin else 0)
        if llim == 0:
            logging.warning("gctrace: possible mismatch of peak start.")
        allpeaks.append({"llim": llim, "rlim": rlim, "max": pmax})
    return allpeaks


def _get_baseline(ys: list[np.ndarray], peakdata: dict) -> list[np.ndarray]:
    """
    Function that returns a baseline based on the trace defined by ``ys``. The
    baseline is created by a linear interpolation between the ``"llim"`` and
    ``"rlim"`` of each `peak` in ``peakdata``. Returns the corrected baseline.
    """
    interpolants = []
    for p in peakdata:
        if len(interpolants) == 0:
            interpolants.append([p["llim"], p["rlim"]])
        else:
            if interpolants[-1][1] == p["llim"]:
                interpolants[-1][1] = p["rlim"]
            else:
                interpolants.append([p["llim"], p["rlim"]])
    bn = ys[0].copy()
    bs = ys[1].copy()
    for pair in interpolants:
        n = pair[1] - pair[0]
        interp = np.interp(range(n), [0, n], [bn[pair[0]], bn[pair[1]]])
        bn[pair[0] : pair[1]] = interp
        bs[pair[0] : pair[1]] = np.zeros(n)
    return bn, bs


def _integrate_peaks(
    xs: list[np.ndarray],
    ys: list[np.ndarray],
    peakdata: dict,
    specdata: dict,
) -> dict:
    """
    A function which, given a trace in [`xs`, `ys`], `peakdata` containing the boundaries
    `"llim"` and `"rlim"` for each `peak`, and `specdata` containing the peak-maximum
    matching limits `"l"` and `"r"`, first assigns peaks into `truepeaks`, then
    baseline-corrects [`xs`, `ys`], and finally integrates the peaks using numpy.trapz().
    """
    xsn = xs[0]
    truepeaks = {}
    for name, species in specdata.items():
        for p in peakdata:
            if xsn[p["max"]] > species["l"] and xsn[p["max"]] < species["r"]:
                truepeaks[name] = p
                break
    bln = _get_baseline(ys, peakdata)
    for k, v in truepeaks.items():
        s = v["llim"]
        e = v["rlim"] + 1
        py = unp.uarray(ys[0][s:e] - bln[0][s:e], ys[1][s:e] - bln[1][s:e])
        px = unp.uarray(xs[0][s:e], xs[1][s:e])
        A = np.trapz(py, px)
        truepeaks[k]["A"] = A
        truepeaks[k]["h"] = py[v["max"] - s]
    return truepeaks


def integrate_trace(traces: dict, chromspec: dict) -> tuple[dict, dict]:
    """
    Integration, calibration, and normalisation handling function. Used to process
    all chromatographic data for which a calibration has been provided

    Parameters
    ----------
    traces
        A dictionary of trace data, with keys being the "raw" name of the detector,
        and the values containing the ``"id"`` for specification matching, and a
        ``"data"`` tuple containing the ``(xs, ys)`` where each element is a pair of
        :class:`(np.ndarray)` with the nominal values and standard deviations.

    chromspec
        Parsed calibration information, with keys being the detector names in the
        calibration file, and values containing the ``"id"`` for detector matching,
        ``"peakdetect"`` dictionary with peak-picking and edge-finding settings,
        and ``"species"`` dictionary with names of species as keys and the left,
        right limits and calibration information as values.

    Returns
    -------
    derived: dict[dict]
        A dictionary containing the derived data, including the peak heights,
        areas, concentrations, mol fractions, and a dictionary with the peak 
        picking information (name, maximum, limits, height, area).
    """

    peaks = {}
    comp = []
    for detname, spec in chromspec.items():
        for det in traces.keys():
            if traces[det]["id"] == spec["id"]:
                traces[det]["calname"] = detname
                break
        units = {
            "t": traces[det]["t"]["u"],
            "y": traces[det]["y"]["u"],
            "A": " ",
        }
        xs, ys = traces[det].pop("data")
        pd = spec.get("peakdetect", {})
        smooth = _get_smooth_yvals(ys[0], pd)
        peakmax, ygrad, _ = _find_peak_maxima(smooth, pd)
        peakspec = _find_peak_edges(smooth, ygrad, peakmax, pd)
        integrated = _integrate_peaks(xs, ys, peakspec, spec["species"])
        peaks[detname] = {}
        for k, v in integrated.items():
            peaks[detname][k] = {
                "peak": {
                    "max": int(v["max"]),
                    "llim": int(v["llim"]),
                    "rlim": int(v["rlim"]),
                },
                "A": {"n": v["A"].n, "s": v["A"].s, "u": units["A"]},
                "h": {"n": v["h"].n, "s": v["h"].s, "u": units["y"]},
            }
            if spec["species"][k].get("calib", None) is not None:
                x = yadg.dgutils.calib_handler(v["A"], spec["species"][k]["calib"])
                x = max(uc.ufloat(0.0, x.s), x)
                peaks[detname][k]["c"] = {
                    "n": x.n,
                    "s": x.s,
                    "u": spec["species"][k]["calib"].get("unit", "%"),
                    "uf": x,
                }
            if k not in comp:
                comp.append(k)
    xout = {}
    area = {}
    conc = {}
    height = {}
    for s in comp:
        for d, ds in chromspec.items():
            if s in peaks[d] and "c" in peaks[d][s]:
                v = peaks[d][s]["c"].pop("uf")
                if ds.get("prefer", False) or s not in xout:
                    xout[s] = v
                    area[s] = peaks[d][s]["A"]
                    height[s] = peaks[d][s]["h"]
                    conc[s] = peaks[d][s]["c"]
    norm = sum([xout[k] for k in xout.keys()])
    for s in xout:
        xnorm = xout[s] / norm
        xout[s] = {"n": xnorm.n, "s": xnorm.s, "u": " "}
    derived = {
        "xout": xout,
        "area": area,
        "height": height,
        "concentration": conc,
        "peaks": peaks,
    }
    return derived



import json
import logging
from uncertainties import ufloat

from yadg.parsers.basiccsv import process_row
import yadg.dgutils

version = "4.0.0"


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    convert: dict = None,
    calfile: str = None,
) -> tuple[list, dict, bool]:
    """
    Legacy MCPT measurement log parser.

    This parser is included to maintain parity with older schemas and datagrams.
    It is essentially a wrapper around :func:`yadg.parsers.basiccsv.process_row`.
    For new applications, please use the `basiccsv` parser.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    convert
        Specification for column conversion. See :ref:`this section<processing_convert>`
        for syntax and further details.

    calfile
        ``convert``-like functionality specified in a json file.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. No metadata is
        returned. The full date is always provided in meascsv-compatible files.

    """
    logging.warning("meascsv: This parser is deprecated. Please switch to 'basiccsv'.")

    if calfile is not None:
        with open(calfile, "r") as infile:
            calib = json.load(infile)
    else:
        calib = {}
    if convert is not None:
        calib.update(convert)

    with open(fn, "r", encoding=encoding) as infile:
        lines = [i.strip() for i in infile.readlines()]

    headers = [i.strip() for i in lines.pop(0).split(";")]
    _units = [i.strip() for i in lines.pop(0).split(";")]
    units = {}
    for h in headers:
        units[h] = _units.pop(0)

    yadg.dgutils.sanitize_units(units)

    datecolumns, datefunc, fulldate = yadg.dgutils.infer_timestamp_from(
        spec={"timestamp": {"index": 0, "format": "%Y-%m-%d-%H-%M-%S"}},
        timezone=timezone,
    )
    timesteps = []
    for line in lines:
        ts = process_row(
            headers, line.split(";"), units, datefunc, datecolumns, calib=calib
        )
        ts["fn"] = str(fn)

        if "derived" in ts:
            xin = {}
            total = ufloat(0, 0)
            for mfc in ts["derived"].keys():
                if mfc.startswith("flow") or mfc.startswith("T"):
                    continue
                xin[mfc] = ufloat(ts["derived"][mfc]["n"], ts["derived"][mfc]["s"])
                total += xin[mfc]
            if xin != {}:
                ts["derived"]["xin"] = {}
                for species in xin:
                    x = xin[species] / total
                    ts["derived"]["xin"][species] = {"n": x.n, "s": x.s, "u": " "}
        timesteps.append(ts)
    return timesteps, None, True



import json
import numpy as np
from operator import itemgetter
from uncertainties import ufloat, umath


def combine(filelist, **kwargs):
    datagramlist = []
    for fn in filelist:
        with open(fn, "r") as infile:
            fndata = json.load(infile)
        for datagram in fndata:
            datagramlist.append(datagram)
    return datagramlist


def reduce(datagramfile, **kwargs):
    tasktype = kwargs.pop("parameters", {"type": "avg"})["type"]
    assert tasktype in [
        "avg",
        "sum",
        "diff",
    ], f'DGUTILS: reduce provided with an incorrent "type": {tasktype}.'
    with open(datagramfile, "r") as infile:
        origdg = json.load(infile)
    if tasktype == "avg":
        weights = [len(origdg) for i in range(len(origdg))]
    elif tasktype == "sum":
        weights = [1 for i in range(len(origdg))]
    elif tasktype == "diff":
        weights = [1] + [-1 for i in range(len(origdg) - 1)]
    dgzero = np.array(origdg[0]["datastream"])
    print(dgzero.shape)
    flatresult = [0.0 for i in dgzero.flatten()]
    for datagram in origdg:
        dgi = np.array(datagram["datastream"])
        assert (
            dgi.shape == dgzero.shape
        ), f"DGUTILS: reduce provided with inconsistent datagram shapes."
        weight = float(weights.pop(0))
        for i in range(len(flatresult)):
            flatresult[i] += dgi.flatten()[i] / weight
    result = {}
    result["datastream"] = np.array(flatresult).reshape(dgzero.shape).tolist()
    result["metadata"] = {}
    result["metadata"]["units"] = origdg[0]["metadata"]["units"]
    result["metadata"]["version"] = kwargs.get("ver", {})
    result["metadata"]["version"]["reduce"] = _VERSION
    return result


def printdg(datagram, **kwargs):
    print(len(datagram))


def pointdata(datagram, pars):
    results = {
        "params": {
            "cavity": {
                "r": pars["cavity"]["r"],
                "h": pars["cavity"]["h"],
                "Q": pars["cavity"]["Q"],
            },
            "sample": {
                "r": pars["sample"]["r"],
                "h": pars["sample"]["h"],
                "rho": pars["sample"]["rho"],
                "m": pars["sample"]["m"],
            },
            "mcpt": {
                "A": pars["mcpt"]["A"],
                "B": pars["mcpt"]["B"],
                "C": pars["mcpt"]["C"],
                "rp_nm": pars["mcpt"]["rp_nm"],
            },
        },
        "metadata": {
            "sample": {
                "name": pars["sample"]["name"],
                "id": pars["sample"]["id"],
                "rep": pars["sample"]["rep"],
                "date": pars["date"],
            }
        },
        "data": [],
    }

    mcptdata = [i for i in datagram if i["input"]["datagram"] == "qftrace"]
    gcdata = [i for i in datagram if i["input"]["datagram"] == "gctrace"]
    expdata = [i for i in datagram if i["input"]["datagram"] == "meascsv"]

    # MCPT parameters
    Vc = (
        np.pi
        * ufloat(*results["params"]["cavity"]["r"]) ** 2
        * ufloat(*results["params"]["cavity"]["h"])
    )
    Vs = (
        np.pi
        * ufloat(*results["params"]["sample"]["r"]) ** 2
        * ufloat(*results["params"]["sample"]["h"])
    )
    delta = (
        ufloat(*results["params"]["sample"]["m"])
        / ufloat(*results["params"]["sample"]["rho"])
    ) * (1 / Vs)
    c_i = pars["cavity"]["i"]
    s_i = pars["sample"]["i"]
    A = pars["mcpt"]["A"]
    B = pars["mcpt"]["B"]
    C = pars["mcpt"]["C"]
    timestamps = []
    for section in mcptdata:
        if pars["reference"] in section["input"]["export"]:
            Q_c = ufloat(
                np.average([p["Q0"][c_i] for p in section["results"][-10:-1]]),
                np.std([p["Q0"][c_i] for p in section["results"][-10:-1]]),
            )
        for p in section["results"]:
            timestamps.append(p["uts"])
    t0 = min(timestamps)
    tinf = max(timestamps) - t0
    Qfac = ufloat(*results["params"]["cavity"]["Q"]["TM020"]) / Q_c
    ffac = ufloat(1 / results["params"]["mcpt"]["rp_nm"], 0)

    results["params"]["cavity"]["Qfac"] = [Qfac.n, Qfac.s, "-"]
    results["params"]["cavity"]["ffac"] = [ffac.n, ffac.s, "-"]

    # inlet flow composition
    xin = []
    for section in expdata:
        xin = xin + [{"uts": p["uts"], "x": p["x"]} for p in section["results"]]
    xin = sorted(xin, key=itemgetter("uts"))

    # process one by one:
    for section in datagram:
        res = []
        if section["input"]["datagram"] == "qftrace":
            for p in section["results"]:
                f0 = p["f0"][c_i] * ffac
                Q0 = p["Q0"][c_i] * Qfac
                fs = p["f0"][s_i] * ufloat(1, 0)
                Qs = p["Q0"][s_i] * ufloat(1, 0)
                r = {
                    "uts": p["uts"],
                    "f0": [f0.n, f0.s, "Hz"],
                    "fs": [fs.n, fs.s, "Hz"],
                    "Q0": [Q0.n, Q0.s, "-"],
                    "Qs": [Qs.n, Qs.s, "-"],
                }
                s = conductivity.Qf2σ(
                    Qs, Q0, fs, f0, Vs, Vc, A=A, B=B, C=C, delta=delta
                )
                for k, v in s.items():
                    r[k] = [v.n, v.s, "S/m" if k == "σ" else "-"]
                res.append(r)
        elif section["input"]["datagram"] == "meascsv":
            for p in section["results"]:
                r = {
                    "uts": p["uts"],
                    "T": [p["T"], 0.05, "°C"],
                    "v·": [p["flow"], 0.005, "ml/min"],
                    "xin": {},
                }
                for key in p["x"]:
                    if key not in r["xin"]:
                        r["xin"][key] = [100 * p["x"][key], 0, "%"]
                tau = Vs / (ufloat(*r["v·"]) * (1e-6 / 60))
                r["τ"] = [tau.n, tau.s, "s"]
                ghsv = ufloat(*r["v·"]) * (1e-6 * 60) / Vs
                r["GHSV"] = [ghsv.n, ghsv.s, "1/h"]
                mvdot = (ufloat(*results["params"]["sample"]["m"]) * (1000)) / (
                    ufloat(*r["v·"]) / 60
                )
                r["m/v·"] = [mvdot.n, mvdot.s, "gs/ml"]
                if "C3H8" in r["xin"]:
                    phi = (ufloat(*r["xin"]["C3H8"]) / ufloat(*r["xin"]["O2"])) / (
                        1 / 5
                    )
                else:
                    phi = ufloat(0)
                r["ϕ"] = [phi.n, phi.s, "-"]
                res.append(r)
        elif section["input"]["datagram"] == "gctrace":
            for p in section["results"]:
                r = {"uts": p["uts"], "xout": {}}
                for key in [
                    "CO",
                    "methane",
                    "CO2",
                    "ethylene",
                    "ethane",
                    "propylene",
                    "propane",
                    "butane",
                    "acetic",
                    "acrylic",
                    "maleic",
                ]:
                    r["xout"][key] = [p["FID"].get(key, {}).get("x", 0), 0, "%"]
                for key in ["O2", "N2"]:
                    r["xout"][key] = [p["TCD"].get(key, {}).get("x", 0), 0, "%"]
                xC3H8 = 100 * np.interp(
                    p["uts"], [xp["uts"] for xp in xin], [xp["x"]["C3H8"] for xp in xin]
                )
                xO2 = 100 * np.interp(
                    p["uts"], [xp["uts"] for xp in xin], [xp["x"]["O2"] for xp in xin]
                )
                if xC3H8 > 1.0:
                    XS = conversion.p2XS(r["xout"], xC3H8, xO2, fuel="propane")
                    for k, v in XS.items():
                        if k.startswith("S"):
                            r[k] = {}
                            for kk, vv in v.items():
                                r[k][kk] = [vv.n, vv.s, "%"]
                        else:
                            r[k] = [v.n, v.s, "%"]
                res.append(r)
        metadata = section.get("metadata", {})
        metadata["dgutils.pointdata"] = {
            "version": _VERSION,
            "date": dateutils.now(asstr=True),
        }
        results["data"].append(
            {"input": section["input"], "metadata": metadata, "results": res}
        )
    return results



import logging
import json
import os
from typing import Union

import yadg.dgutils
import yadg.core

version = "4.0.0"


def calib_3to4(oldcal: dict, caltype: str) -> dict:
    newcal = {}
    if caltype == "calfile":
        for k, v in oldcal["detectors"].items():
            pd = {
                "window": (v.get("window", 3) - 1) // 2,
                "polyorder": v.get("poly", 2),
                "prominence": v.get("prominence", 1.0),
                "threshold": v.get("threshold", 1.0),
            }
            sp = {}
            for kk, vv in v["species"].items():
                if kk == "units" and vv == "min":
                    continue
                spec = {
                    "l": vv["l"] * 60.0,
                    "r": vv["r"] * 60.0,
                    "calib": {"inverse": {"slope": vv.get("rf", 1.0)}},
                }
                sp[kk] = spec
            id = {"det_1": 0, "det_2": 1}[v["id"]]
            newcal[k] = {"id": id, "peakdetect": pd, "species": sp}
    elif caltype == "Tcalfile":
        newcal = {"T": {"T_f": {"calib": {"linear": oldcal}}, "unit": "degC"}}
    elif caltype == "MFCcalfile":
        newcal = {}
        for k, v in oldcal.items():
            items = v.get("content", {k: 1.0})
            for kk, vv in items.items():
                if kk not in newcal:
                    newcal[kk] = {"unit": "ml/min"}
                newcal[kk][k] = {
                    "calib": {
                        "linear": {
                            "slope": v.get("slope", 1.0),
                            "intercept": v.get("intercept", 0.0),
                        }
                    },
                    "fraction": vv,
                }
    return newcal


def schema_3to4(oldschema: list) -> dict:
    newschema = {
        "metadata": {
            "provenance": {
                "yadg": yadg.dgutils.get_yadg_metadata(),
                "update_schema": {"updater": "schema_3to4"},
            },
            "schema_version": version,
            "timezone": "localtime",
        },
        "steps": [],
    }
    for oldstep in oldschema:
        newstep = {}

        newstep["parser"] = oldstep["datagram"]
        if newstep["parser"] == "gctrace":
            newstep["parser"] = "chromtrace"

        if "paths" in oldstep["import"]:
            oldstep["import"]["files"] = oldstep["import"].pop("paths")
        newstep["import"] = oldstep["import"]

        if oldstep.get("export", None) is not None:
            newstep["tag"] = oldstep["export"]

        calib = {}
        species = {}
        detectors = {}
        parameters = {}
        for k, v in oldstep["parameters"].items():
            if k in ["Tcalfile", "MFCcalfile", "calfile"]:
                try:
                    with open(v, "r") as cf:
                        temp = calib_3to4(json.load(cf), k)
                    logging.info(
                        f"schema_3to4: Found calfile '{k}' at '{v}', "
                        f"will attempt to modernise the calibration."
                    )
                    if k == "calfile":
                        for kk, vv in temp.items():
                            species[kk] = vv.pop("species")
                            detectors[kk] = vv
                    else:
                        calib.update(temp)
                except IOError:
                    logging.warning(
                        f"schema_3to4: Error reading '{k}' file: '{v}'. "
                        f"Keeping original key-value pair."
                    )
                    parameters[k] = v
            elif k == "method" and v == "q0refl":
                v = "kajfez"
                parameters[k] = v
            else:
                parameters[k] = v
        if newstep["parser"] == "meascsv" and "flow" not in calib:
            calib["flow"] = {"flow high": {}, "flow low": {}, "unit": "ml/min"}
        if calib != {}:
            parameters["convert"] = calib
        if species != {}:
            parameters["species"] = species
        if detectors != {}:
            parameters["detectors"] = detectors
        if parameters != {}:
            newstep["parameters"] = parameters
        newschema["steps"].append(newstep)

    return newschema


def datagram_3to4(olddg: list) -> dict:
    logging.warning(
        "datagram_3to4: updating from datagram version 3.1.0 is lossy. It is "
        "considerably safer to update schema instead, and re-process the raw data."
    )

    newdg = {
        "metadata": {
            "provenance": {
                "yadg": yadg.dgutils.get_yadg_metadata(),
                "update_object": {"updater": "datagram_3to4", "version": version},
            },
            "datagram_version": version,
            "date": yadg.dgutils.now(asstr=True),
            "input_schema": {},
        },
        "steps": [],
    }

    oldschema = []
    for oldstep in olddg:
        step = {}
        oldschema.append(oldstep["input"])
        step["metadata"] = {
            "tag": oldstep["input"]["export"],
            "parser": oldstep["metadata"],
        }
        step["metadata"]["parser"]["update_object"] = {
            "version": version,
            "updater": "datagram_3to4",
        }

        step["data"] = []
        for timestep in oldstep["results"]:
            ts = {"raw": {}, "derived": {}}
            ts["uts"] = float(timestep.pop("uts"))

            if oldstep["input"]["datagram"] == "meascsv":
                if "files" in oldstep["input"]["import"]:
                    ts["fn"] = oldstep["input"]["import"]["files"][0]
                else:
                    ts["fn"] = oldstep["input"]["import"]["paths"][0]
            elif oldstep["input"]["datagram"] == "qftrace":
                ts["fn"] = timestep.pop("path")
            elif oldstep["input"]["datagram"] == "gctrace":
                ts["fn"] = timestep.pop("fn")

            for k, v in timestep.items():
                if oldstep["input"]["datagram"] == "meascsv":
                    if k == "raw":
                        for kk, vv in v.items():
                            if isinstance(vv, (int, str)):
                                ts["raw"][kk] = vv
                            else:
                                if kk.startswith("T"):
                                    s = 0.1
                                    u = "%" if kk.endswith("o") else "degC"
                                elif kk == "pressure":
                                    s = 0.1
                                    u = "mbar"
                                elif kk == "heater flow":
                                    s = 0.1
                                    u = "l/min"
                                else:
                                    s = 0.1
                                    u = "ml/min"
                                ts["raw"][kk] = {"n": vv, "s": s, "u": u}
                    elif k == "T":
                        ts["derived"]["T"] = {"n": v, "s": 5.0, "u": "degC"}
                    elif k == "flow":
                        ts["derived"]["flow"] = {"n": v, "s": 0.001, "u": "ml/min"}
                    elif k == "x":
                        ts["derived"]["xin"] = {}
                        for kk, vv in v.items():
                            ts["derived"]["xin"][kk] = {"n": vv, "s": 0.001, "u": " "}

                elif oldstep["input"]["datagram"] == "qftrace":
                    if k in ["Q0", "f0"]:
                        ts["derived"][k[0]] = {
                            "n": v,
                            "s": [20.0 if k == "Q0" else 1000.0] * len(v),
                            "u": " " if k == "Q0" else "Hz",
                        }

                elif oldstep["input"]["datagram"] == "gctrace":
                    if "peaks" not in ts["derived"]:
                        ts["derived"]["peaks"] = {}
                    if "xout" not in ts["derived"]:
                        ts["derived"]["xout"] = {}
                    if k in ["TCD", "FID"]:
                        ts["derived"]["peaks"][k] = {}
                        for kk, vv in v.items():
                            ts["derived"]["peaks"][k][kk] = {
                                "A": {"n": vv["A"], "s": 0.01 * vv["A"], "u": " "},
                                "h": {"n": vv["h"], "s": 1.0, "u": " "},
                                "c": {"n": vv["x"], "s": 0.01 * vv["x"], "u": "mol/l"},
                            }
                            if kk not in ts["derived"]["xout"] or k == "FID":
                                ts["derived"]["xout"][kk] = {
                                    "n": vv["x"] / 100,
                                    "s": 0.01 * vv["x"] / 100,
                                    "u": " ",
                                }
            if oldstep["input"]["datagram"] == "gctrace":
                totx = sum([v["n"] for k, v in ts["derived"]["xout"].items()])
                for k in ts["derived"]["xout"].keys():
                    ts["derived"]["xout"][k]["n"] = ts["derived"]["xout"][k]["n"] / totx
                    ts["derived"]["xout"][k]["s"] = ts["derived"]["xout"][k]["s"] / totx
            step["data"].append(ts)
        newdg["steps"].append(step)
    newdg["metadata"]["input_schema"] = schema_3to4(oldschema)

    return newdg


def update_object(type: str, object: Union[list, dict]) -> dict:
    """
    Yadg's update worker function.

    This is the main function called when **yadg** is executed as ``yadg update``.
    The main idea is to allow a simple update pathway from older versions of `schema` and
    ``datagram`` files to the current latest and greatest.

    Currently supports:

     - updating ``schema`` version 3.1.0 to 4.0.0
     - updating ``datagram`` version 3.1.0 to 4.0.0, although you **really** should be
       updating the ``schema`` instead and re-processing files.

    Parameters
    ----------
    type
        The type of the passed object, either `"datagram"` or `"schema"`.

    object
        The object to be updated

    Returns
    -------
    newobj: dict
        The updated and validated `"datagram"` or `"schema"`.

    """
    assert type in [
        "datagram",
        "schema",
    ], f"update_object: Provided type '{type}' is not one of ['datagram', 'schema']. "

    oldver = None
    if isinstance(object, list):
        oldver = "3.1.0"

    assert oldver is not None, (
        f"update_object: Impossible to figure out old '{type}' version. "
        "Please convert manually."
    )

    # distribute to updaters
    if oldver.startswith("3.") and type == "schema":
        logging.info("update_object: Updating old schema.")
        newobj = schema_3to4(object)
    elif oldver.startswith("3.") and type == "datagram":
        logging.info("update_object: Updating old datagram.")
        newobj = datagram_3to4(object)
    elif oldver.startswith("4."):
        logging.info("update_object: Already at latest version, no update necessary.")
        newobj = object

    with open("temp.json", "w") as outfile:
        json.dump(newobj, outfile, indent=1)

    if type == "schema":
        logging.info("update_object: Validating new schema.")
        yadg.core.validators.validate_schema(
            newobj, strictfiles=False, strictfolders=False
        )
    elif type == "datagram":
        logging.info("update_object: Validating new datagram.")
        yadg.core.validators.validate_datagram(newobj)

    return newobj


def schema_from_preset(preset: dict, folder: str) -> dict:
    newmeta = {
        "provenance": {
            "yadg": yadg.dgutils.get_yadg_metadata(),
        },
        "schema_version": version,
    }
    preset["metadata"].update(newmeta)
    for step in preset["steps"]:
        k = "files" if "files" in step["import"] else "folders"
        newf = []
        for oldf in step["import"][k]:
            if os.path.isabs(oldf):
                logging.warning(
                    f"schema_from_preset: Item '{oldf}' in '{k}' is an absolute "
                    "path and will not be patched."
                )
            else:
                assert not oldf.startswith("." + os.path.sep), (
                    f"schema_from_preset: Item '{oldf}' in '{k}' does start "
                    f"with '.{os.path.sep}' and therefore should not be patched "
                    f"using '{folder}'."
                )
                newp = os.path.abspath(os.path.join(folder, oldf))
                newf.append(newp)
        step["import"][k] = newf
        if "calfile" in step["parameters"]:
            oldf = step["parameters"]["calfile"]
            if os.path.isabs(oldf):
                logging.warning(
                    f"schema_from_preset: Specified calfile '{oldf}' is an absolute "
                    "path and will not be patched."
                )
            else:
                newp = os.path.abspath(os.path.join(folder, oldf))
                step["parameters"]["calfile"] = newp
    return preset



import yadg.dgutils

version = "4.0.0"


def process(
    fn: str, encoding: str = "utf-8", timezone: str = "localtime", **kwargs: dict
) -> tuple[list, dict, bool]:
    """
    A dummy parser.

    This parser simply returns the current time, the filename provided, and any
    ``kwargs`` passed.

    Parameters
    ----------
    fn
        Filename to process

    encoding
        Not used.

    timezone
        Not used

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. No metadata is
        returned by the dummy parser. The full date is always returned.

    """

    result = {"uts": yadg.dgutils.now(), "fn": str(fn), "raw": kwargs}

    return [result], None, True
